@@ -5,6 +5,7 @@
 
 #include "mined.h"
 #include "charcode.h"
+#include "termprop.h"
 
 
 /*======================================================================*\
@@ -22,7 +23,7 @@ FLAG
 no_unichar (u)
   unsigned long u;
 {
-  return u == 0 || no_char (u);
+  return u == CHAR_UNKNOWN || u == CHAR_INVALID;
 }
 
 /**
@@ -31,12 +32,22 @@ no_unichar (u)
  */
 int
 iscontrol (c)
-  character c;
+  unsigned long c;
 {
   if (mapped_text) {
 	unsigned long u = lookup_encodedchar (c);
 	return u == '\177' || (! no_unichar (u) && u < ' ');
-  } else if (utf8_text || cjk_text) {
+  } else if (utf8_text) {
+	if (unassigned_single_width) {
+		if (rxvt_version > 0) {
+			/* handle weird mapping of non-Unicode ranges */
+			if (c < 0x80000000) {
+				c &= 0x1FFFFF;
+			}
+		}
+	}
+	return c == '\177' || c < ' ';
+  } else if (cjk_text) {
 	return c == '\177' || c < ' ';
   } else {
 	return c == '\177' || (c & '\177') < ' ';
@@ -391,10 +402,10 @@ struct charmap_table_entry {
 static struct charmap_table_entry charmaps_table [] =
 {
 #ifdef __TURBOC__
-	{pc437_table, & pc437_table_len, "CP437", "PC", 'p'},
-	{pc850_table, & pc850_table_len, "CP850", "PL", 'P'},
+	{cp437_table, & cp437_table_len, "CP437", "PC", 'p'},
+	{cp850_table, & cp850_table_len, "CP850", "PL", 'P'},
 #else
-#include "charmaps.c"
+#include "charmaps.t"
 #endif
 };
 
@@ -405,7 +416,8 @@ static struct charmap_table_entry charmaps_table [] =
 
 /**
    matchprefix determines whether its first parameter contains its 
-   second parameter matching approximately as an initial prefix.
+   second parameter matching approximately as an initial prefix,
+   at word boundaries!
    The match ignores separating '-', '_', and space characters, 
    and does not match case.
    The algorithm assumes that letters are ASCII as this is used for 
@@ -427,6 +439,17 @@ matchprefix (s, m)
 	}
 	if (! * m) {
 		return True;
+#ifdef koi8_ru_fix
+		/* approx. prefix match found; check word boundary */
+		if ( (* s >= 'a' && * s <= 'z')
+		  || (* s >= 'A' && * s <= 'Z')
+		  || (* s >= '0' && * s <= '9')
+		   ) {
+			/* continue */
+		} else {
+			return True;
+		}
+#endif
 	}
 	if (! * s) {
 		return False;
@@ -450,7 +473,8 @@ matchprefix (s, m)
 /**
    matchpart determines whether its first parameter contains its 
    second parameter matching approximately as an initial prefix or 
-   as a prefix of any part after a '/' or '>' separator.
+   as a prefix of any part after a '/' or '>' separator,
+   at word boundaries!
    The match ignores separating '-', '_', and space characters, 
    and does not match case.
    The algorithm assumes that letters are ASCII as this is used for 
@@ -562,9 +586,9 @@ get_text_encoding ()
 char *
 get_term_encoding ()
 {
-  if (* term_encoding == '\0' && utf8_screen) {
+  if (utf8_screen) {
 	return "UTF-8";
-  } else if (* term_encoding == '\0' && ! cjk_term && ! mapped_term) {
+  } else if (! cjk_term && ! mapped_term) {
 	return "ISO 8859-1";
   } else {
 	return term_encoding;
@@ -604,7 +628,7 @@ mapping_has_combining (term)
 	} else {
 		unichar = lookup_encodedchar (i);
 	}
-	if (iscombining_char (unichar)) {
+	if (term ? term_iscombining (unichar) : iscombining_unichar (unichar)) {
 		return True;
 	}
   }
@@ -748,7 +772,10 @@ set_char_encoding (term, charmap, tag)
 	if (term) {
 		utf8_screen = True;
 		utf8_input = True;
+		cjk_term = False;
+		mapped_term = False;
 		term_encoding = "UTF-8";
+		term_encoding_tag = 'U';
 	} else {
 		utf8_text = True;
 		utf16_file = False;
@@ -763,7 +790,9 @@ set_char_encoding (term, charmap, tag)
 		utf8_screen = False;
 		utf8_input = False;
 		cjk_term = False;
+		mapped_term = False;
 		term_encoding = "ISO 8859-1";
+		term_encoding_tag = 'L';
 	} else {
 		utf8_text = False;
 		cjk_text = False;
@@ -811,6 +840,8 @@ set_term_encoding (charmap, tag)
   return set_char_encoding (True, charmap, tag);
 }
 
+#define dont_debug_set_text_encoding
+
 /**
    Set character mapping table and text encoding variables according 
    to encoding tag.
@@ -822,7 +853,9 @@ set_text_encoding (charmap, tag, debug_tag)
   char tag;
   char * debug_tag;
 {
-/*printf ("set_text_encoding [%s] %s %c\n", debug_tag, charmap, tag);*/
+#ifdef debug_set_text_encoding
+  printf ("set_text_encoding [%s] %s %c\n", debug_tag, charmap, tag);
+#endif
   return set_char_encoding (False, charmap, tag);
 }
 
@@ -908,7 +941,7 @@ map_char (cjk, map_table, map_table_len)
 			low = i + 1;
 		}
 	}
-	return 0;
+	return CHAR_INVALID;
 }
 
 
@@ -928,7 +961,7 @@ gb_to_unicode (gb)
 	unsigned int byte3 = (gb >> 8) & 0xFF;
 	unsigned int byte4 = gb & 0xFF;
 	if (byte2 < '0' || byte2 > '9' || byte3 < 0x81 || byte4 < '0' || byte4 > '9') {
-		return 0;
+		return CHAR_INVALID;
 	}
 	return (((((gb >> 24) & 0xFF) - 0x90) * 10
 		+ (byte2 - 0x30)) * 126L
@@ -1068,7 +1101,7 @@ lookup_mapped_char (cjk, map_table, map_table_len)
 		return cjk;
 	} else {
 		/* notify "not found" */
-		return 0;
+		return CHAR_INVALID;
 	}
 }
 
