@@ -16,7 +16,7 @@ struct hanentry hantable [] =
 
 #ifndef __EMX__	/* EMX cannot compile this ("virtual memory exhausted") */
 
-#include "handescr.h"
+#include "handescr.t"
 
 #endif
 
