@@ -1,3 +1,11 @@
+/* options */
+#define dont_dynamic_hilite_tracking
+#define dont_debug_hilite_tracking
+#define dont_debug_mouse
+extern void mouse_hilite_mode _((FLAG hilite));
+extern void mouse_hilite_tracking_on _((int startx, int starty, int firstrow, int lastrow));
+extern void mouse_hilite_tracking_off _((void));
+extern void mouse_button_event_mode _((FLAG report));
 
 /* signal handling */
 extern void catch_signals _((void (* catchfunc) (void)));
@@ -11,10 +19,17 @@ extern void getwinsize _((void));
 
 /* i/o handling */
 extern void flush _((void));
-extern int inputreadyafter _((int fid, int msec));
 extern void putescape _((char * s));
 
-/* screen mode handling */
+/* i/o internal */
+extern int inputreadyafter _((int fid, int msec));
+
+/* terminal encoding detection */
+#ifdef msdos
+extern int get_codepage _((void));
+#endif
+
+/* terminal screen mode handling */
 #ifdef msdos
 extern void set_video_lines _((int r));
 	/* 0/1/2: 200/350/400 lines */
@@ -34,24 +49,32 @@ extern void switch_textmode_height _((FLAG cycle));
 	/* True: cycle through font heights 8/14/16 
 	   False: switch between font heights 8/16 */
 
-/* screen manipulation */
+/* terminal screen manipulation */
 extern void clear_screen _((void));
 extern void clear_eol _((void));
+extern void erase_chars _((int));
 extern void scroll_forward _((void));
 extern void scroll_reverse _((void));
 extern void add_line _((int));
 extern void delete_line _((int));
 extern void set_cursor _((int, int));
 extern int current_cursor_y;
+#define in_status_line	(current_cursor_y == YMAX)
 
 /* screen output */
+extern void clearscreen _((void));
 extern void putstring _((char *));
 extern void __putchar _((character));
 extern void ring_bell _((void));
 extern void putblockchar _((character));
 
+/* modal output */
+extern void put_menu_marker _((void));
+
 /* screen attributes: ANSI sequences */
-extern char * dimansi;		/* line indicator dimming */
+extern char * markansi;		/* line indicator marking (dim/red) */
+extern char * emphansi;		/* status line emphasis mode (red bg) */
+extern char * borderansi;	/* menu border colour (red) */
 extern char * selansi;		/* menu selection */
 extern char * selfgansi;	/* menu selection background */
 extern char * ctrlansi;		/* control character display */
@@ -72,8 +95,9 @@ extern void reverse_on _((void));
 extern void reverse_off _((void));
 extern void altcset_on _((void));
 extern void altcset_off _((void));
-extern void dim_on _((void));
-extern void dim_off _((void));
+extern void emph_on _((void));
+extern void mark_on _((void));
+extern void mark_off _((void));
 extern void bold_on _((void));
 extern void bold_off _((void));
 extern void unidisp_on _((void));
@@ -101,6 +125,10 @@ extern void menuborder_on _((void));
 extern void menuborder_off _((void));
 extern void menuitem_on _((void));
 extern void menuitem_off _((void));
+extern void set_colour_token _((int));
+
+extern char restore_ansi7 [];
+extern FLAG do_restore_ansi7;
 
 /* input functions */
 extern int get_digits _((int *));
@@ -117,6 +145,7 @@ struct pc_fkeyentry {
 	unsigned char fkeyshift;
 };
 extern struct pc_fkeyentry pc_fkey_map [256];
+extern struct pc_fkeyentry pc_xkey_map [512];
 struct fkeyentry {
 	char * fk;
 	voidfunc fp;
@@ -132,7 +161,7 @@ extern unsigned long readcharacter_unicode _((void));
 extern unsigned long readcharacter_unicode_mapped _((void));
 extern unsigned long _readchar_nokeymap _((void));
 extern int read1byte _((void));
-extern int char_ready_within _((int));
+extern int char_ready_within _((int msec, char * debug_tag));
 
 /* mouse handling functions */
 extern void DIRECTcrttool _((void));
@@ -151,6 +180,10 @@ extern FLAG can_scroll_reverse;
 extern FLAG can_add_line;
 extern FLAG can_delete_line;
 extern FLAG can_clear_eol;
+extern FLAG can_erase_chars;
+extern FLAG can_reverse_mode;
+extern FLAG can_hide_cursor;
+extern FLAG can_dim;
 extern FLAG can_alt_cset;
 
 /* terminal feature usage flags */
@@ -164,10 +197,14 @@ extern int menumargin;
 extern FLAG use_stylish_menu_selection;
 extern int use_unicode_menubar _((void));
 extern FLAG explicit_border_style;
+extern FLAG explicit_scrollbar_style;
+extern FLAG use_script_colour;
 extern FLAG use_mouse;
 extern FLAG use_bold;
+extern FLAG avoid_reverse_colour;
 extern FLAG use_bgcolor;
 extern FLAG use_modifyOtherKeys;
+extern FLAG blink_cursor;
 extern FLAG colours_256;
 extern FLAG colours_88;
 extern FLAG ansi_esc;
@@ -205,7 +242,7 @@ ctrl	alt	shift	ESC code/ -1	keyshift code
 #define altctrl_mask	(alt_mask | ctrl_mask)
 #define altctrlshift_mask	(alt_mask | ctrl_mask | shift_mask)
 #define applkeypad_mask	0x08
-/* skip 0x30 (may be included from keyboard.c) */
+/* skip 0x30 (may be left in from keyboard.c) */
 #define report_mask	0x40
 
 extern int ansi_params;
