@@ -5,6 +5,7 @@
 
 #include "mined.h"
 #include "io.h"
+#include "termprop.h"
 
 
 #ifndef __TURBOC__
@@ -72,7 +73,8 @@ static unsigned long _readchar _((void));
 |*		Handling function key and keyboard mapping tables	*|
 \*======================================================================*/
 
-#define in_status_line	(current_cursor_y == YMAX)
+#define dont_debug_findkey
+
 
 #ifndef msdos
 
@@ -85,6 +87,7 @@ dummyfunc ()
 extern struct fkeyentry fkeymap [];
 extern struct fkeyentry fkeymap_xterm [];
 extern struct fkeyentry fkeymap_rxvt [];
+extern struct fkeyentry fkeymap_linux [];
 extern struct fkeyentry fkeymap_vt100 [];
 extern struct fkeyentry fkeymap_vt52 [];
 extern struct fkeyentry fkeymap_hp [];
@@ -92,32 +95,38 @@ extern struct fkeyentry fkeymap_siemens [];
 extern struct fkeyentry fkeymap_scoansi [];
 extern struct fkeyentry fkeymap_interix [];
 
-struct fkeyentry * fkeymap2 = fkeymap_xterm;
+static struct fkeyentry * fkeymap_spec = fkeymap_xterm;
+extern struct fkeyentry fkeymap_terminfo [];
 
 void
 set_fkeymap (term)
   char * term;
 {
+#ifdef debug_findkey
+  printf ("set_fkeymap %s\n", term);
+#endif
   if (term == NIL_PTR) {
-	return;
+	fkeymap_spec = fkeymap;
   } else if (* term == 'x') {
-	fkeymap2 = fkeymap_xterm;
+	fkeymap_spec = fkeymap_xterm;
   } else if (* term == 'r') {
-	fkeymap2 = fkeymap_rxvt;
+	fkeymap_spec = fkeymap_rxvt;
+  } else if (* term == 'l') {
+	fkeymap_spec = fkeymap_linux;
   } else if (* term == 'h') {
-	fkeymap2 = fkeymap_hp;
+	fkeymap_spec = fkeymap_hp;
 	hp_shift_mode = True;
   } else if (* term == 'v') {
-	fkeymap2 = fkeymap_vt100;
+	fkeymap_spec = fkeymap_vt100;
   } else if (* term == '5') {
-	fkeymap2 = fkeymap_vt52;
+	fkeymap_spec = fkeymap_vt52;
   } else if (* term == 'i') {
-	fkeymap2 = fkeymap_interix;
+	fkeymap_spec = fkeymap_interix;
   } else if (* term == 's') {
-	fkeymap2 = fkeymap_siemens;
+	fkeymap_spec = fkeymap_siemens;
 	use_mouse = False;
   } else if (* term == 'o') {
-	fkeymap2 = fkeymap_scoansi;
+	fkeymap_spec = fkeymap_scoansi;
 	sco_shift_mode = True;
 	use_mouse = False;
   }
@@ -161,12 +170,12 @@ notify_input_method ()
 	case 'X':	source = "X"; break;
   }
   if (source) {
-	build_string (text_buffer, "Input method: %s (source: %s)", 
+	build_string (text_buffer, "Input method: %s (source: %s)", 
 			km->itemname, source);
   } else {
-	build_string (text_buffer, "Input method: %s", km->itemname);
+	build_string (text_buffer, "Input method: %s", km->itemname);
   }
-  status_msg (text_buffer);
+  status_uni (text_buffer);
 }
 
 /**
@@ -185,7 +194,7 @@ set_keymap (new_keymap, current, next)
   keyboard_mapping = current;
   next_keyboard_mapping = next;
 
-  if (! loading && ! in_status_line) {
+  if (! loading && ! in_status_line && ! input_active) {
 	notify_input_method ();
   }
 }
@@ -366,6 +375,10 @@ get_digits (result)
 |*				Terminal input				*|
 \*======================================================================*/
 
+#define dont_debug_mapped_keyboard
+
+#define dont_debug_focus
+
 #define max_stamps 10
 static long last_sec = 0;
 static long last_msec = 0;
@@ -376,6 +389,10 @@ static long total_deltas = 0;
 long average_delta_readchar = 0;
 long last_delta_readchar = 0;
 
+static FLAG skip_1click = False;
+static FLAG skip_1release = False;
+
+
 static
 void
 timestamp_readchar ()
@@ -450,6 +467,9 @@ readchar ()
   while (command (c) == MOUSEescape
 	&& mouse_button == releasebutton && report_release == False)
   {
+#ifdef debug_focus
+	printf ("ignoring MOUSE release event\n");
+#endif
 	c = _readchar ();
   }
   report_release = False;
@@ -514,8 +534,6 @@ readcharacter_allbuttons ()
   return c;
 }
 
-#define dont_debug_mapped_keyboard
-
 /*
    readcharacter_mapping () reads one character
    either with or without keyboard mapping
@@ -541,8 +559,66 @@ readcharacter_mapping (map_keyboard, return_unicode)
   }
 
   if (unichar == FUNcmd) {
-	/* pass back WINCH exception */
-	return FUNcmd;
+	/** housekeeping of some events to cover the cases:
+		focus-out mouse-click focus-in mouse-release (xterm 224)
+		focus-out mouse-click mouse-release focus-in
+		focus-out focus-in mouse-click mouse-release
+	    and ignore the first mouse click/release in any case
+	 */
+	FLAG skipmouse = False;
+
+	if (command (unichar) == FOCUSout) {
+#ifdef debug_focus
+		printf ("FOCUS out\n");
+#endif
+		FOCUSout ();
+		skip_1click = True;
+		skip_1release = True;
+		skipmouse = True;
+	} else if (command (unichar) == FOCUSin) {
+#ifdef debug_focus
+		printf ("FOCUS in\n");
+#endif
+		FOCUSin ();
+		skipmouse = True;
+	} else if (command (unichar) == MOUSEescape) {
+		if (mouse_button == releasebutton && skip_1release) {
+#ifdef debug_focus
+			printf ("skipping 1 MOUSE release event\n");
+#endif
+			skipmouse = True;
+			skip_1release = False;
+		} else if (mouse_button != releasebutton && skip_1click) {
+#ifdef debug_focus
+			printf ("skipping 1 MOUSE event\n");
+#endif
+			skipmouse = True;
+			skip_1click = False;
+			skip_1release = False;
+		} else if (! window_focus) {
+			/* should not be needed anymore */
+#ifdef debug_focus
+			printf ("skipping MOUSE event while focus out\n");
+#endif
+			skipmouse = True;
+		}
+	}
+
+	if (skipmouse) {
+		return readcharacter_mapping (map_keyboard, return_unicode);
+	} else {
+		/* WINCH exception isn't passing by here... */
+#ifdef debug_focus
+		if (command (unichar) == MOUSEescape) {
+			if (mouse_button == releasebutton) {
+				printf ("passing MOUSE release event\n");
+			} else {
+				printf ("passing MOUSE event\n");
+			}
+		}
+#endif
+		return FUNcmd;
+	}
   }
 
   if (utf8_input || queue_mode_mapped) {
@@ -598,9 +674,6 @@ readcharacter_mapping (map_keyboard, return_unicode)
 	}
 	if (return_unicode || utf8_text) {
 		unichar = lookup_encodedchar (unichar);
-		if (no_unichar (unichar)) {
-			return CHAR_INVALID;
-		}
 	}
 	return unichar;
   } else {
@@ -692,7 +765,6 @@ DIRECTxtermgetxy (code)
   character button;
   int xpos;
   int ypos;
-  char track [22];	/* for mouse hilite tracking */
 
   mouse_lastbutton = mouse_button;
   if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
@@ -751,6 +823,31 @@ DIRECTxtermgetxy (code)
   }
   mouse_ypos = ypos - 33 - MENU;
 
+#ifdef debug_hilite_tracking
+  printf ("mouse %c: ", code);
+  {
+	static char * button_name [] = {
+		"releasebutton",
+		"leftbutton",
+		"middlebutton",
+		"rightbutton",
+		"movebutton",
+		"wheelup",
+		"wheeldown"};
+	if (mouse_button != releasebutton) {
+		printf ("%s", button_name [mouse_button]);
+	} else {
+		printf ("release (%s)", button_name [mouse_prevbutton]);
+	}
+  }
+  printf (" @ %d, %d\n", mouse_xpos, mouse_ypos);
+#endif
+
+#ifdef unix
+  if (mouse_button_event_tracking && mouse_button == releasebutton) {
+	mouse_button_event_mode (False);
+  }
+
   if (mouse_hilite_tracking && mouse_button == leftbutton) {
 	/* handle hilite tracking */
 	if (mouse_ypos < 0 || mouse_ypos > last_y 
@@ -758,16 +855,25 @@ DIRECTxtermgetxy (code)
 	    || is_menu_open ())
 	{
 		/* abort hilite tracking */
-		putescape ("\033[0;0;0;0;0T");
-		flush ();
+		mouse_hilite_tracking_off ();
+
+#ifdef dynamic_hilite_tracking
+		/* disable hilite tracking mode */
+		mouse_hilite_mode (False);
+#endif
 	} else {
-		build_string (track, "\033[1;%d;%d;%d;%dT", 
-				mouse_xpos + 1, mouse_ypos + 1 + MENU, 
-				1 + MENU, last_y + 3);
-		putescape (track);
-		flush ();
+#ifdef dynamic_hilite_tracking
+		/* enable hilite tracking mode */
+		mouse_hilite_mode (True);
+#endif
+
+		mouse_hilite_tracking_on (mouse_xpos + 1, 
+					  mouse_ypos + 1 + MENU, 
+					  1 + MENU, 
+					  last_y + 3);
 	}
   }
+#endif
 }
 
 static
@@ -813,12 +919,14 @@ DIRECTxterm ()
   DIRECTxtermgetxy ('M');
   MOUSEescape ();
 }
+
 void
 TRACKxterm ()
 {
   DIRECTxtermgetxy ('t');
   MOUSEescape ();
 }
+
 void
 TRACKxtermT ()
 {
@@ -1162,12 +1270,29 @@ findkey (str)
   char * str;
 {
   int dummy;
-  int res = findkeyin (str, fkeymap2, 0, & dummy);
-  if (res == -2) {
-	return findkeyin (str, fkeymap, 0, & dummy);
-  } else {
+  int res;
+
+  res = findkeyin (str, fkeymap_spec, 0, & dummy);
+#ifdef debug_findkey
+  printf ("<^[%s> _spec %d\n", str + 1, res);
+#endif
+  if (res != -2) {
 	return res;
   }
+
+  res = findkeyin (str, fkeymap_terminfo, 0, & dummy);
+#ifdef debug_findkey
+  printf ("<^[%s> _terminfo %d\n", str + 1, res);
+#endif
+  if (res != -2) {
+	return res;
+  }
+
+  res = findkeyin (str, fkeymap, 0, & dummy);
+#ifdef debug_findkey
+  printf ("<^[%s> _ %d\n", str + 1, res);
+#endif
+  return res;
 }
 
 /*
@@ -1271,7 +1396,8 @@ mapkb (str, kbmap, matchmode, found, mapped)
 			* found = kbmap [i].fk;
 			* mapped = kbmap [i].fp;
 			if (! * mapped) {
-				* mapped = I;
+				/* fix missing mapping in keyboard mapping table */
+				* mapped = "";
 			}
 			if (matchmode == 1) {
 				/* return index unless other prefix 
@@ -1547,33 +1673,66 @@ read1byte ()
   }
 }
 
+
+#define dont_debug_terminal_response
+#ifdef debug_terminal_response
+#define trace_terminal_response(what)	if (debug_tag) printf ("[%s (%d)] %s\n", debug_tag, msec, what);
+#else
+#define trace_terminal_response(what)
+#endif
+
 #ifndef msdos
 
 static
 int
-bytereadyafter (msec)
+bytereadyafter (msec, debug_tag)
   int msec;
+  char * debug_tag;
 {
-  return have_rest_queue || inputreadyafter (input_fd, msec);
+  if (have_rest_queue) {
+	trace_terminal_response ("from rest queue");
+	return 1;
+  } else {
+	if (inputreadyafter (input_fd, msec) > 0) {
+		trace_terminal_response ("response");
+		return 1;
+	} else {
+		trace_terminal_response ("timeout");
+		return 0;
+	}
+  }
 }
 
 /*
  * Is a character available within a specified number of milliseconds ?
  */
 int
-char_ready_within (msec)
+char_ready_within (msec, debug_tag)
   int msec;
+  char * debug_tag;
 {
-  return (q_len () > 0) || bytereadyafter (msec);
+  if (q_len () > 0) {
+	trace_terminal_response ("from queue");
+	return 1;
+  } else {
+	return bytereadyafter (msec, debug_tag);
+  }
 }
 
 #else	/* ifndef msdos */
 
 int
-char_ready_within (msec)
+char_ready_within (msec, debug_tag)
   int msec;
+  char * debug_tag;
 {
-  return inputreadyafter (input_fd, msec);
+  if (inputreadyafter (input_fd, msec) > 0) {
+	trace_terminal_response ("response");
+	return 1;
+  } else {
+	trace_terminal_response ("timeout");
+	return 0;
+  }
 }
 
 #endif
@@ -1588,60 +1747,61 @@ char_ready_within (msec)
 static menuitemtype radicalsmenu [] =
 {
 	/* 1 stroke radicals */
-	{"①-1:一 2:丨 3:丶 4:丿乀乁 5:乙乚乛 6:亅", dummyfunc, ""},
+	{"①.1:一 2:丨 3:丶 4:丿乀乁 5:乙乚乛 6:亅", dummyfunc, ""},
 	/* 2 strokes radicals */
-	{"②-7:二 8:亠 9:人亻 10:儿 11:入 12:八 13:冂 14:冖", dummyfunc, ""},
-	{"②-15:冫 16:几 17:凵 18:刀刁刂 19:力 20:勹 21:匕", dummyfunc, ""},
-	{"②-22:匚 23:匸 24:十 25:卜 26:卩 27:厂 28:厶 29:又", dummyfunc, ""},
+	{"②.7:二 8:亠 9:人亻 10:儿 11:入 12:八 13:冂 14:冖", dummyfunc, ""},
+	{"②.15:冫 16:几 17:凵 18:刀刁刂 19:力 20:勹 21:匕", dummyfunc, ""},
+	{"②.22:匚 23:匸 24:十 25:卜 26:卩 27:厂 28:厶 29:又", dummyfunc, ""},
 	/* 3 strokes radicals */
-	{"③-30:口 31:囗 32:土 33:士 34:夂 35:夊 36:夕 37:大夨 38:女 39:子孑孒孓", dummyfunc, ""},
-	{"③-40:宀 41:寸 42:小 43:尢尣 44:尸 45:屮 46:山 47:巛巜川 48:工 49:己已巳", dummyfunc, ""},
-	{"③-50:巾 51:干 52:乡幺 53:广 54:廴 55:廾 56:弋 57:弓 58:彐彑 59:彡 60:彳", dummyfunc, ""},
+	{"③.30:口 31:囗 32:土 33:士 34:夂 35:夊 36:夕 37:大夨 38:女 39:子孑孒孓", dummyfunc, ""},
+	{"③.40:宀 41:寸 42:小 43:尢尣 44:尸 45:屮 46:山 47:巛巜川 48:工 49:己已巳", dummyfunc, ""},
+	{"③.50:巾 51:干 52:乡幺 53:广 54:廴 55:廾 56:弋 57:弓 58:彐彑 59:彡 60:彳", dummyfunc, ""},
 	/* 4 strokes radicals */
-	{"④-61:心忄 62:戈 63:戶户戸 64:手扌才 65:支 66:攴攵 67:文 68:斗", dummyfunc, ""},
-	{"④-69:斤 70:方 71:无 72:日 73:曰 74:月 75:木朩 76:欠 77:止 78:歹 79:殳", dummyfunc, ""},
-	{"④-80:毋毌 81:比 82:毛 83:氏 84:气 85:水氵 86:火灬 87:爪爫 88:父 89:爻", dummyfunc, ""},
-	{"④-90:丬爿 91:片 92:牙㸦 93:牛牜 94:犬犭", dummyfunc, ""},
+	{"④.61:心忄 62:戈 63:戶户戸 64:手扌才 65:支 66:攴攵 67:文 68:斗", dummyfunc, ""},
+	{"④.69:斤 70:方 71:无 72:日 73:曰 74:月 75:木朩 76:欠 77:止 78:歹 79:殳", dummyfunc, ""},
+	{"④.80:毋毌 81:比 82:毛 83:氏 84:气 85:水氵 86:火灬 87:爪爫 88:父 89:爻", dummyfunc, ""},
+	{"④.90:丬爿 91:片 92:牙㸦 93:牛牜 94:犬犭", dummyfunc, ""},
 	/* 5 strokes radicals */
-	{"⑤-95:玄 96:玉王 97:瓜 98:瓦 99:甘 100:生 101:用甩 102:田由甲申甴电", dummyfunc, ""},
-	{"⑤-103:疋 104:疒 105:癶 106:白 107:皮 108:皿 109:目", dummyfunc, ""},
-	{"⑤-110:矛 111:矢 112:石 113:示礻 114:禸 115:禾 116:穴 117:立", dummyfunc, ""},
+	{"⑤.95:玄 96:玉王 97:瓜 98:瓦 99:甘 100:生 101:用甩 102:田由甲申甴电", dummyfunc, ""},
+	{"⑤.103:疋 104:疒 105:癶 106:白 107:皮 108:皿 109:目", dummyfunc, ""},
+	{"⑤.110:矛 111:矢 112:石 113:示礻 114:禸 115:禾 116:穴 117:立", dummyfunc, ""},
 	/* 6 strokes radicals */
-	{"⑥-118:竹 119:米 120:糸糹纟 121:缶 122:网罒罓䍏 123:羊 124:羽 125:老耂考", dummyfunc, ""},
-	{"⑥-126:而 127:耒 128:耳 129:聿肀 130:肉 131:臣 132:自", dummyfunc, ""},
-	{"⑥-133:至 134:臼 135:舌 136:舛 137:舟 138:艮 139:色 140:艸艹䒑", dummyfunc, ""},
-	{"⑥-141:虍 142:虫 143:血 144:行 145:衣衤 146:襾西覀", dummyfunc, ""},
+	{"⑥.118:竹 119:米 120:糸糹纟 121:缶 122:网罒罓䍏 123:羊 124:羽 125:老耂考", dummyfunc, ""},
+	{"⑥.126:而 127:耒 128:耳 129:聿肀 130:肉 131:臣 132:自", dummyfunc, ""},
+	{"⑥.133:至 134:臼 135:舌 136:舛 137:舟 138:艮 139:色 140:艸艹䒑", dummyfunc, ""},
+	{"⑥.141:虍 142:虫 143:血 144:行 145:衣衤 146:襾西覀", dummyfunc, ""},
 	/* 7 strokes radicals */
-	{"⑦-147:見见 148:角 149:言訁讠 150:谷 151:豆 152:豕 153:豸", dummyfunc, ""},
-	{"⑦-154:貝贝 155:赤 156:走赱 157:足 158:身 159:車车 160:辛", dummyfunc, ""},
-	{"⑦-161:辰 162:辵辶 163:邑 164:酉 165:釆 166:里", dummyfunc, ""},
+	{"⑦.147:見见 148:角 149:言訁讠 150:谷 151:豆 152:豕 153:豸", dummyfunc, ""},
+	{"⑦.154:貝贝 155:赤 156:走赱 157:足 158:身 159:車车 160:辛", dummyfunc, ""},
+	{"⑦.161:辰 162:辵辶 163:邑 164:酉 165:釆 166:里", dummyfunc, ""},
 	/* 8 strokes radicals */
-	{"⑧-167:金釒钅 168:長镸长 169:門门 170:阜阝", dummyfunc, ""},
-	{"⑧-171:隶 172:隹 173:雨 174:靑青 175:非", dummyfunc, ""},
+	{"⑧.167:金釒钅 168:長镸长 169:門门 170:阜阝", dummyfunc, ""},
+	{"⑧.171:隶 172:隹 173:雨 174:靑青 175:非", dummyfunc, ""},
 	/* 9 strokes radicals */
-	{"⑨-176:面靣 177:革 178:韋韦 179:韭 180:音 181:頁页", dummyfunc, ""},
-	{"⑨-182:風风 183:飛飞 184:食飠饣 185:首 186:香", dummyfunc, ""},
+	{"⑨.176:面靣 177:革 178:韋韦 179:韭 180:音 181:頁页", dummyfunc, ""},
+	{"⑨.182:風风 183:飛飞 184:食飠饣 185:首 186:香", dummyfunc, ""},
 	/* 10 strokes radicals */
-	{"⑩-187:馬马 188:骨 189:高髙 190:髟 191:鬥 192:鬯 193:鬲 194:鬼", dummyfunc, ""},
+	{"⑩.187:馬马 188:骨 189:高髙 190:髟 191:鬥 192:鬯 193:鬲 194:鬼", dummyfunc, ""},
 	/* 11 strokes radicals */
-	{"⑪-195:魚鱼 196:鳥鸟 197:鹵 198:鹿 199:麥麦 200:麻", dummyfunc, ""},
+	{"⑪.195:魚鱼 196:鳥鸟 197:鹵 198:鹿 199:麥麦 200:麻", dummyfunc, ""},
 	/* 12 strokes radicals */
-	{"⑫-201:黃黄 202:黍 203:黑黒 204:黹", dummyfunc, ""},
+	{"⑫.201:黃黄 202:黍 203:黑黒 204:黹", dummyfunc, ""},
 	/* 13 strokes radicals */
-	{"⑬-205:黽黾 206:鼎 207:鼓鼔 208:鼠鼡", dummyfunc, ""},
+	{"⑬.205:黽黾 206:鼎 207:鼓鼔 208:鼠鼡", dummyfunc, ""},
 	/* 14 strokes radicals */
-	{"⑭-209:鼻 210:齊齐", dummyfunc, ""},
+	{"⑭.209:鼻 210:齊齐", dummyfunc, ""},
 	/* 15 strokes radicals */
-	{"⑮-211:齒齿", dummyfunc, ""},
+	{"⑮.211:齒齿", dummyfunc, ""},
 	/* 16 strokes radicals */
-	{"⑯-212:龍龙 213:龜龟", dummyfunc, ""},
+	{"⑯.212:龍龙 213:龜龟", dummyfunc, ""},
 	/* 17 strokes radical */
-	{"⑰-214:龠", dummyfunc, ""},
+	{"⑰.214:龠", dummyfunc, ""},
 };
 #endif
 
 
 #define dont_debug_kbesc
+#define dont_debug_kbansi
 #define dont_debug_kbmap
 
 static
@@ -1801,6 +1961,15 @@ _readchar ()
 
 #ifdef msdos
 
+  if (ch > 0xFF) {
+	keyproc = pc_xkey_map [ch - 0x100].fp;
+	if (! keyproc) {
+		keyproc = I;
+	}
+	keyshift = pc_xkey_map [ch - 0x100].fkeyshift;
+	return FUNcmd;
+  }
+
   if (ch == '\000') {
 # ifdef dosmouse
 	ch = getch ();
@@ -1833,20 +2002,23 @@ _readchar ()
 
 	while  ((res = findkey (ctrl_queue)) == -1 /* prefix of table entry */
 		&& q_notfull ()
-		&& bytereadyafter (escape_delay)
+		&& bytereadyafter (escape_delay, NIL_PTR)
 	       )
 	{
 		char prev_ch = ch;
 		ch = read1byte ();
 #ifdef debug_kbesc
-		printf ("<^[%s> + %02X\n", ctrl_queue + 1, ch);
+		printf ("<^[%s> + [%dms] %02X\n", ctrl_queue + 1, escape_delay, ch);
 #endif
 
 		/* accept ESC prefix as Alt modifier */
 		if (detect_esc_alt && ch == '\033' && prev_ch == '\033') {
-			if (bytereadyafter (escape_delay)) {
+			if (bytereadyafter (escape_delay, NIL_PTR)) {
 				ch = read1byte ();
 				keyshift |= alt_mask;
+#ifdef debug_kbesc
+				printf ("-> <^[%s> + ^[/%02X\n", ctrl_queue + 1, ch);
+#endif
 			}
 		}
 
@@ -1859,21 +2031,38 @@ _readchar ()
 				keyshift |= report_mask;
 			}
 			ch = read1byte ();
+#ifdef debug_kbesc
+			printf ("-> <^[%s> + >?/%02X\n", ctrl_queue + 1, ch);
+#endif
 		}
 
 check_ctrl_byte:
 		/* handle generic shift state of ANSI sequences: */
+		/* syntax of modifier indication varies with 
+		   terminal type and mode; indications are as follows:
+					ANSI 1	ANSI 2	keyshift
+		   plain		-	9?	0x00
+		   Shift		2	10	0x01
+		   Alt			3	11	0x02
+		   Alt-Shift		4	12	0x03
+		   Control		5	13	0x04
+		   Control-Shift	6	14	0x05
+		   Alt-Control		7	15	0x06
+		   Alt-Control-Shift	8	16	0x07
+		   applkeypad_mask			0x08
+		 */
 		if (ch == ';' && second_esc_char == '[') {
 			ch = read1byte ();
-			if (ch >= '2' && ch <= '8') {
+			if (ch >= '2' && ch <= '9') {
 				/* ESC [1;2A -> ESC [1A + shift etc */
+				/* special: map '9' to applkeypad_mask */
 				keyshift |= (ch & ~'0') - 1;
 			} else if (ch == '1') {
 				ch = read1byte ();
 				if (ch >= '0' && ch <= '6') {
 					/* xterm with alwaysUseMods: */
 					/* ESC [1;11A -> ESC [1A + alt etc */
-					keyshift |= applkeypad_mask | ((ch & ~'0')  + 1);
+					keyshift |= applkeypad_mask | ((ch & ~'0') + 1);
 				} else {
 					q_put (';');
 					q_put ('1');
@@ -1884,29 +2073,36 @@ check_ctrl_byte:
 				q_put (ch);
 			}
 			ch = read1byte ();
+#ifdef debug_kbesc
+			printf ("-> <^[%s> + 1-8/%02X\n", ctrl_queue + 1, ch);
+#endif
 		} else if (ch == ';' && second_esc_char == '1') {
 			ch = read1byte ();
 			if (ch >= '2' && ch <= '8') {
 				/* xterm/VT52 mode: ESC 1;2A -> ESC 1A + shift etc */
-				keyshift |= (ch & ~'0')  - 1;
+				keyshift |= (ch & ~'0') - 1;
 			} else {
 				q_put (';');
 				q_put (ch);
 			}
 			ch = read1byte ();
+#ifdef debug_kbesc
+			printf ("-> <^[%s> + ;/%02X\n", ctrl_queue + 1, ch);
+#endif
 		} else if (ch >= '0' && ch <= '9'
 			   && (second_esc_char == 'O' || 
 			       (hp_shift_mode && ! second_esc_char))) {
-			if (ch >= '2' && ch <= '8') {
+			if (ch >= '2' && ch <= '9') {
 				/* ESC O2A -> ESC OA + shift etc */
 				/* HP: ESC 2p -> ESC p + shift etc */
-				keyshift |= (ch & ~'0')  - 1;
+				/* special: map '9' to applkeypad_mask */
+				keyshift |= (ch & ~'0') - 1;
 			} else if (ch == '1') {
 				ch = read1byte ();
 				if (ch >= '0' && ch <= '6') {
 					/* ESC O11P -> ESC OP + alt etc */
 					/* HP: ESC 11p -> ESC p + alt etc */
-					keyshift |= applkeypad_mask | ((ch & ~'0')  + 1);
+					keyshift |= applkeypad_mask | ((ch & ~'0') + 1);
 				} else {
 					q_put ('1');
 					q_put (ch);
@@ -1915,16 +2111,20 @@ check_ctrl_byte:
 				q_put (ch);
 			}
 			ch = read1byte ();
+#ifdef debug_kbesc
+			printf ("-> <^[%s> + 0-9/%02X\n", ctrl_queue + 1, ch);
+#endif
 		} else if (sco_shift_mode && second_esc_char == '['
 			   && ch >= '0' && ch <= '9') {
 			char newkeyshift;
 			char newch = '\0';
-			if (ch >= '2' && ch <= '8') {
+			if (ch >= '2' && ch <= '9') {
 				/* SCO: ESC [2M -> ESC [M + shift etc */
-				newkeyshift = (ch & ~'0')  - 1;
+				/* special: map '9' to applkeypad_mask */
+				newkeyshift = (ch & ~'0') - 1;
 				newch = read1byte ();
 #ifdef debug_kbesc
-		printf ("-> <^[%s> + 2-8/%02X\n", ctrl_queue + 1, newch);
+				printf ("-> <^[%s> + 2-8/%02X\n", ctrl_queue + 1, newch);
 #endif
 				if (newch >= 'A' && newch <= 'Z') {
 					keyshift |= newkeyshift;
@@ -1934,11 +2134,11 @@ check_ctrl_byte:
 			} else if (ch == '1') {
 				ch = read1byte ();
 #ifdef debug_kbesc
-		printf ("-> <^[%s> + 1/%02X\n", ctrl_queue + 1, ch);
+				printf ("-> <^[%s> + 1/%02X\n", ctrl_queue + 1, ch);
 #endif
 				if (ch >= '0' && ch <= '6') {
 					/* SCO: ESC [11M -> ESC [M + alt etc */
-					newkeyshift = applkeypad_mask | ((ch & ~'0')  + 1);
+					newkeyshift = applkeypad_mask | ((ch & ~'0') + 1);
 					newch = read1byte ();
 					if (newch >= 'A' && newch <= 'Z') {
 						keyshift |= newkeyshift;
@@ -1958,7 +2158,7 @@ check_ctrl_byte:
 				}
 			} else {
 #ifdef debug_kbesc
-		printf ("-> <^[%s>\n", ctrl_queue + 1);
+				printf ("-> <^[%s>\n", ctrl_queue + 1);
 #endif
 				q_put (ch);
 				newch = read1byte ();
@@ -2008,12 +2208,16 @@ check_ctrl_byte:
 		if (second_esc_char == '\0') {
 			second_esc_char = ch;
 			if (second_esc_char == 'O') {
+				/* appl. cursor keys, appl. keypad */
 				keyshift |= applkeypad_mask;
 			}
 		}
 	}
 #ifdef debug_kbesc
-	printf ("=> <^[%s> [%X]\n", ctrl_queue + 1, keyshift);
+	printf ("=> [%d] <^[%s> [keyshift %X]\n", res, ctrl_queue + 1, keyshift);
+#endif
+#ifdef debug_kbansi
+	printf ("[%d] [keyshift %X]\n", res, keyshift);
 #endif
 
 	if (quit) {
@@ -2036,18 +2240,45 @@ check_ctrl_byte:
 				}
 			} while (ansi_fini == ';');
 			q_clear ();
+#ifdef debug_kbansi
+			printf ("ansi %c (%d)", ansi_fini, ansi_params);
+			{int i;
+			 for (i = 0; i < ansi_params; i ++) {
+				printf (" %d", ansi_param [i]);
+			 }
+			}
+			printf ("\n");
+#endif
 
-			if (ansi_params >= 3
-			    && ansi_param [0] == 27
-			    && ansi_fini == '~') {
-				/* handle generic escape sequence 
-				   for modified key 
-				   (xterm 214 resource modifyOtherKeys)
+			if ((ansi_fini == '~'
+			     && ansi_param [0] == 27
+			     && ansi_params >= 3
+			    )
+			 || (ansi_fini == 'u'
+			     && (ansi_params == 2
+			      || (ansi_params == 1 && keyshift > 0)
+			        )
+			    )
+			   ) {
+				/* map xterm 235 mode formatOtherKeys: 1
+					^[[77;2u
+				   to xterm mode formatOtherKeys: 0
+					^[[27;2;77~
 				 */
-				if (ansi_param [1] > 8) {
-					ansi_param [1] -= 8;
+				if (ansi_fini == 'u') {
+					ansi_param [2] = ansi_param [0];
+				}
+
+				if (ansi_params > 1) {
+					/* handle generic escape sequence 
+					   for modified key 
+					   (xterm 214 resource modifyOtherKeys)
+					 */
+					if (ansi_param [1] > 8) {
+						ansi_param [1] -= 8;
+					}
+					keyshift |= ansi_param [1] - 1;
 				}
-				keyshift |= ansi_param [1] - 1;
 
 				/* enforce UTF-8 interpretation of input */
 				queue_mode_mapped = True;
@@ -2083,7 +2314,7 @@ check_ctrl_byte:
 							int len;
 							utf8_info (ps->pat1, & len, & unichar);
 							ansi_param [2] = unichar;
-						} else if ('?' <= ansi_param [2] && ansi_param [2] <= '') {
+						} else if ('?' <= ansi_param [2] && ansi_param [2] < '') {
 							keyshift &= ~ ctrl_mask;
 							return ansi_param [2] & 0x1F;
 						}
@@ -2152,12 +2383,21 @@ check_ctrl_byte:
 		keyshift |= fkeyshift;
 
 		if (keyproc == DIRECTxterm) {
+#ifdef debug_mouse
+			printf ("xterm mouse\n");
+#endif
 			DIRECTxtermgetxy ('M');
 			keyproc = MOUSEescape;
 		} else if (keyproc == TRACKxterm) {
+#ifdef debug_mouse
+			printf ("xterm mouse tracking\n");
+#endif
 			DIRECTxtermgetxy ('t');
 			keyproc = MOUSEescape;
 		} else if (keyproc == TRACKxtermT) {
+#ifdef debug_mouse
+			printf ("xterm mouse tracking\n");
+#endif
 			DIRECTxtermgetxy ('T');
 			keyproc = MOUSEescape;
 		} else if (keyproc == DIRECTcrttool) {
@@ -2176,10 +2416,18 @@ check_ctrl_byte:
 	while  ((res = mapkb (ctrl_queue, keyboard_map, 1, & found, & mapped)) == -1
 		 /* prefix of some table entry */
 		&& q_notfull ()
-		&& bytereadyafter (keymap_delay)
+		&& bytereadyafter (keymap_delay, NIL_PTR)
 	       )
 	{
+#ifdef fix_modified_input
+		/* avoid xterm modifyOtherKeys mode to interfere with 
+		   input method handling: will have to use separate queue
+		 */
+		q_put (_readchar_nokeymap ());
+		plainly like this, we get an endless loop...
+#else
 		q_put (read1byte ());
+#endif
 		/* exact match found? note for later consideration */
 		if (found != NIL_PTR) {
 			prev_found = found;
