@@ -7,13 +7,6 @@
 #include "io.h"
 
 
-/*======================================================================*\
-|*	Dispatch function keys and some keypad keys to mined functions	*|
-\*======================================================================*/
-
-#include "dispatch.c"
-
-
 /*======================================================================*\
 |*			Key and function key -> command mapping		*|
 \*======================================================================*/
@@ -24,10 +17,14 @@
 
 voidfunc key_map [256] =
 {	/* characters to functions map */
-  /* ^@-^G */	MARK, MPW, DCC, PD, MRT, MUP, MNW, GOTO,
-  /* ^H-^O */	DPC, Sch, SNL, DLN, HOP, SNL, RS, LIB,
-  /* ^P-^W */	PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
-  /* ^X-^_ */	MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
+  /* ^@ ^A ^B ^C ^D ^E ^F ^G */
+		MARK, MPW, DCC, PD, MRT, MUP, MNW, GOTO,
+  /* ^H ^I ^J ^K ^L ^M ^N ^O */
+		DPC, Sch, Sch, DLN, HOP, SNL, RS, LIB,
+  /* ^P ^Q ^R ^S ^T ^U ^V ^W */
+		PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
+  /* ^X ^Y ^Z ^[ ^\ ^] ^^ ^_ */
+		MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
   /* 040-057 */	Sch, Sch, Sdoublequote, Sch, Sch, Sch, Sch, Ssinglequote,
 		Sch, Sch, Sch, Sch, Sch, Sdash, Sch, Sch,
   /* 060-077 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
@@ -56,22 +53,22 @@ voidfunc ws_key_map [32] =
 
 voidfunc pico_key_map [32] =
 {	/* characters to functions map */
-  /* ^@-^G */	MNW, BLINE, MLF, FS, DCC, ELINE, MRT, HELP,
-  /* ^H-^O */	DPC, Sch, JUS, DLN, RDwin, SNL, MDNEL, WTU,
+  /* ^@-^G */	MNW, GO_LINEBEGIN , MLF, FS, DCC, GO_LINEEND , MRT, HELP,
+  /* ^H-^O */	DPC, Sch, JUS, DLN, RDwin, SNL, MDN, WTU,
   /* ^P-^W */	MUP, I, INSFILE, I, SPELL, PASTE, PD, SFW,
   /* ^X-^_ */	QUED, PU, I, ESCAPE, CANCEL, MARK, toggleMARK, CTRLINS,
 };
 
 voidfunc emacs_key_map [32] =
 {	/* characters to functions map */
-  /* ^@-^G */	MARK, BLINE, MLF, I, DCC, ELINE, MRT, CANCEL,
+  /* ^@-^G */	MARK, GO_LINEBEGIN , MLF, I, DCC, GO_LINEEND , MRT, CANCEL,
   /* ^H-^O */	HELP, Sch, Sch, DLN, RD, SNL, MDN, LIB,
   /* ^P-^W */	MUP, CTRLINS, SRV, SFW, I, (voidfunc) REPT, PD, CUT,
   /* ^X-^_ */	EMAX, PASTE, SUSP, META, HOP, I, HOP, UNDO,
 };
 
 
-/* mapping function key codes -> functions
+/* mapping DOS 2-byte (leading 0) function key codes -> functions
    function keys deliver NUL and the following codes:
 block1:	RGISOQHKPM
 block2:	GHIKMOPQRS	Pos1 Up PgUp Left Center Right End Down PdDn Ins
@@ -231,6 +228,216 @@ struct pc_fkeyentry pc_fkey_map [256] =
    /* F0-FF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 };
 
+/* mapping DOS extended function key codes -> functions
+   keys as delivered by djgpp function getxkey, listed in keys.h
+*/
+struct pc_fkeyentry pc_xkey_map [512] =
+{
+					{I},
+/* #define K_Alt_Escape           0x101 */	{EXED},
+					{I},
+/* #define K_Control_At           0x103 */	{MARK},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+/* #define K_Alt_Backspace        0x10e */	{DPC, ctrl_mask},
+/* #define K_BackTab              0x10f */	{MPW},
+/* #define K_Alt_Q                0x110 */	{I},
+/* #define K_Alt_W                0x111 */	{I},
+/* #define K_Alt_E                0x112 */	{EDITMENU},
+/* #define K_Alt_R                0x113 */	{I},
+/* #define K_Alt_T                0x114 */	{I},
+/* #define K_Alt_Y                0x115 */	{I},
+/* #define K_Alt_U                0x116 */	{I},
+/* #define K_Alt_I                0x117 */	{I},
+/* #define K_Alt_O                0x118 */	{OPTIONSMENU},
+/* #define K_Alt_P                0x119 */	{PARAMENU},
+/* #define K_Alt_LBracket         0x11a */	{I},
+/* #define K_Alt_RBracket         0x11b */	{I},
+/* #define K_Alt_Return           0x11c */	{SNL, alt_mask},
+					{I},
+/* #define K_Alt_A                0x11e */	{I},
+/* #define K_Alt_S                0x11f */	{SEARCHMENU},
+/* #define K_Alt_D                0x120 */	{I},
+/* #define K_Alt_F                0x121 */	{FILEMENU},
+/* #define K_Alt_G                0x122 */	{I},
+/* #define K_Alt_H                0x123 */	{I},
+/* #define K_Alt_J                0x124 */	{I},
+/* #define K_Alt_K                0x125 */	{I},
+/* #define K_Alt_L                0x126 */	{I},
+/* #define K_Alt_Semicolon        0x127 */	{I},
+/* #define K_Alt_Quote            0x128 */	{I},
+/* #define K_Alt_Backquote        0x129 */	{I},
+					{I},
+/* #define K_Alt_Backslash        0x12b */	{I},
+/* #define K_Alt_Z                0x12c */	{I},
+/* #define K_Alt_X                0x12d */	{I},
+/* #define K_Alt_C                0x12e */	{I},
+/* #define K_Alt_V                0x12f */	{I},
+/* #define K_Alt_B                0x130 */	{I},
+/* #define K_Alt_N                0x131 */	{I},
+/* #define K_Alt_M                0x132 */	{I},
+/* #define K_Alt_Comma            0x133 */	{I},
+/* #define K_Alt_Period           0x134 */	{I},
+/* #define K_Alt_Slash            0x135 */	{I},
+					{I},
+/* #define K_Alt_KPStar           0x137 */	{I},
+					{I}, {I}, {I},
+/* #define K_F1                   0x13b */	{F1},
+/* #define K_F2                   0x13c */	{F2},
+/* #define K_F3                   0x13d */	{F3},
+/* #define K_F4                   0x13e */	{F4},
+/* #define K_F5                   0x13f */	{F5},
+/* #define K_F6                   0x140 */	{F6},
+/* #define K_F7                   0x141 */	{F7},
+/* #define K_F8                   0x142 */	{F8},
+/* #define K_F9                   0x143 */	{F9},
+/* #define K_F10                  0x144 */	{F10},
+					{I}, {I},
+/* #define K_Home                 0x147 */	{HOMEkey},
+/* #define K_Up                   0x148 */	{MUP},
+/* #define K_PageUp               0x149 */	{PU},
+/* #define K_Alt_KPMinus          0x14a */	{I},
+/* #define K_Left                 0x14b */	{MLF},
+/* #define K_Center               0x14c */	{HOP},
+/* #define K_Right                0x14d */	{MRT},
+/* #define K_Alt_KPPlus           0x14e */	{I},
+/* #define K_End                  0x14f */	{ENDkey},
+/* #define K_Down                 0x150 */	{MDN},
+/* #define K_PageDown             0x151 */	{PD},
+/* #define K_Insert               0x152 */	{INSkey},
+/* #define K_Delete               0x153 */	{DELkey},
+/* #define K_Shift_F1             0x154 */	{F1, shift_mask},
+/* #define K_Shift_F2             0x155 */	{F2, shift_mask},
+/* #define K_Shift_F3             0x156 */	{F3, shift_mask},
+/* #define K_Shift_F4             0x157 */	{F4, shift_mask},
+/* #define K_Shift_F5             0x158 */	{F5, shift_mask},
+/* #define K_Shift_F6             0x159 */	{F6, shift_mask},
+/* #define K_Shift_F7             0x15a */	{F7, shift_mask},
+/* #define K_Shift_F8             0x15b */	{F8, shift_mask},
+/* #define K_Shift_F9             0x15c */	{F9, shift_mask},
+/* #define K_Shift_F10            0x15d */	{F10, shift_mask},
+/* #define K_Control_F1           0x15e */	{F1, ctrl_mask},
+/* #define K_Control_F2           0x15f */	{F2, ctrl_mask},
+/* #define K_Control_F3           0x160 */	{F3, ctrl_mask},
+/* #define K_Control_F4           0x161 */	{F4, ctrl_mask},
+/* #define K_Control_F5           0x162 */	{F5, ctrl_mask},
+/* #define K_Control_F6           0x163 */	{F6, ctrl_mask},
+/* #define K_Control_F7           0x164 */	{F7, ctrl_mask},
+/* #define K_Control_F8           0x165 */	{F8, ctrl_mask},
+/* #define K_Control_F9           0x166 */	{F9, ctrl_mask},
+/* #define K_Control_F10          0x167 */	{F10, ctrl_mask},
+/* #define K_Alt_F1               0x168 */	{F1, alt_mask},
+/* #define K_Alt_F2               0x169 */	{F2, alt_mask},
+/* #define K_Alt_F3               0x16a */	{F3, alt_mask},
+/* #define K_Alt_F4               0x16b */	{F4, alt_mask},
+/* #define K_Alt_F5               0x16c */	{F5, alt_mask},
+/* #define K_Alt_F6               0x16d */	{F6, alt_mask},
+/* #define K_Alt_F7               0x16e */	{F7, alt_mask},
+/* #define K_Alt_F8               0x16f */	{F8, alt_mask},
+/* #define K_Alt_F9               0x170 */	{F9, alt_mask},
+/* #define K_Alt_F10              0x171 */	{F10, alt_mask},
+/* #define K_Control_Print        0x172 */	{I},
+/* #define K_Control_Left         0x173 */	{MLF, ctrl_mask},
+/* #define K_Control_Right        0x174 */	{MRT, ctrl_mask},
+/* #define K_Control_End          0x175 */	{ENDkey, ctrl_mask},
+/* #define K_Control_PageDown     0x176 */	{PD, ctrl_mask},
+/* #define K_Control_Home         0x177 */	{HOMEkey, ctrl_mask},
+/* #define K_Alt_1                0x178 */	{I},
+/* #define K_Alt_2                0x179 */	{I},
+/* #define K_Alt_3                0x17a */	{I},
+/* #define K_Alt_4                0x17b */	{I},
+/* #define K_Alt_5                0x17c */	{I},
+/* #define K_Alt_6                0x17d */	{I},
+/* #define K_Alt_7                0x17e */	{I},
+/* #define K_Alt_8                0x17f */	{I},
+/* #define K_Alt_9                0x180 */	{I},
+/* #define K_Alt_0                0x181 */	{I},
+/* #define K_Alt_Dash             0x182 */	{I},
+/* #define K_Alt_Equals           0x183 */	{I},
+/* #define K_Control_PageUp       0x184 */	{PU, ctrl_mask},
+/* #define K_F11                  0x185 */	{F11},
+/* #define K_F12                  0x186 */	{F12},
+/* #define K_Shift_F11            0x187 */	{F11, shift_mask},
+/* #define K_Shift_F12            0x188 */	{F12, shift_mask},
+/* #define K_Control_F11          0x189 */	{F11, ctrl_mask},
+/* #define K_Control_F12          0x18a */	{F12, ctrl_mask},
+/* #define K_Alt_F11              0x18b */	{F11, alt_mask},
+/* #define K_Alt_F12              0x18c */	{F12, alt_mask},
+/* #define K_Control_Up           0x18d */	{MUP, ctrl_mask},
+/* #define K_Control_KPDash       0x18e */	{I},
+/* #define K_Control_Center       0x18f */	{HOP, ctrl_mask},
+/* #define K_Control_KPPlus       0x190 */	{I},
+/* #define K_Control_Down         0x191 */	{MDN, ctrl_mask},
+/* #define K_Control_Insert       0x192 */	{INSkey, ctrl_mask},
+/* #define K_Control_Delete       0x193 */	{DELkey, ctrl_mask},
+					{I},
+/* #define K_Control_KPSlash      0x195 */	{I},
+/* #define K_Control_KPStar       0x196 */	{I},
+/* #define K_Alt_EHome            0x197 */	{smallHOMEkey, alt_mask},
+/* #define K_Alt_EUp              0x198 */	{MUP, alt_mask},
+/* #define K_Alt_EPageUp          0x199 */	{PU, alt_mask},
+					{I},
+/* #define K_Alt_ELeft            0x19b */	{MLF, alt_mask},
+					{I},
+/* #define K_Alt_ERight           0x19d */	{MRT, alt_mask},
+					{I},
+/* #define K_Alt_EEnd             0x19f */	{smallENDkey, alt_mask},
+/* #define K_Alt_EDown            0x1a0 */	{MDN, alt_mask},
+/* #define K_Alt_EPageDown        0x1a1 */	{PD, alt_mask},
+/* #define K_Alt_EInsert          0x1a2 */	{smallINSkey, alt_mask},
+/* #define K_Alt_EDelete          0x1a3 */	{smallDELkey, alt_mask},
+/* #define K_Alt_KPSlash          0x1a4 */	{I},
+/* #define K_Alt_Tab              0x1a5 */	{I},
+/* #define K_Alt_Enter            0x1a6 */	{SNL, alt_mask},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I},
+/* #define K_EHome                0x247 */	{smallHOMEkey},
+/* #define K_EUp                  0x248 */	{MUP},
+/* #define K_EPageUp              0x249 */	{PU},
+					{I},
+/* #define K_ELeft                0x24b */	{MLF},
+					{I},
+/* #define K_ERight               0x24d */	{MRT},
+					{I},
+/* #define K_EEnd                 0x24f */	{smallENDkey},
+/* #define K_EDown                0x250 */	{MDN},
+/* #define K_EPageDown            0x251 */	{PD},
+/* #define K_EInsert              0x252 */	{smallINSkey},
+/* #define K_EDelete              0x253 */	{smallDELkey},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+					{I}, {I}, {I},
+/* #define K_Control_ELeft        0x273 */	{MLF, ctrl_mask},
+/* #define K_Control_ERight       0x274 */	{MRT, ctrl_mask},
+/* #define K_Control_EEnd         0x275 */	{smallENDkey, ctrl_mask},
+/* #define K_Control_EPageDown    0x276 */	{PD, ctrl_mask},
+/* #define K_Control_EHome        0x277 */	{smallHOMEkey, ctrl_mask},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+					{I}, {I}, {I}, {I},
+/* #define K_Control_EPageUp      0x284 */	{PU, ctrl_mask},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+/* #define K_Control_EUp          0x28d */	{MUP, ctrl_mask},
+					{I}, {I}, {I},
+/* #define K_Control_EDown        0x291 */	{MDN, ctrl_mask},
+/* #define K_Control_EInsert      0x292 */	{smallINSkey, ctrl_mask},
+/* #define K_Control_EDelete      0x293 */	{smallDELkey, ctrl_mask},
+					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+};
 
 #ifndef msdos
 
@@ -264,18 +471,18 @@ struct fkeyentry fkeymap [] =
 	{"\033[219z",	MRT},	/* more cursor right	*/
 	{"\033[217z",	MLF},	/* more cursor left	*/
 
-	/* SuSe Linux console */
-	{"\033[1~",	smallHOMEkey},	/* Pos1	*/
-	{"\033[4~",	smallENDkey},	/* End	*/
-
-	/* RedHat Linux console */
+	/* Linux console, Cygwin console */
+	{"\033[1~",	smallHOMEkey},	/* Pos1, VT220: FIND (@0)	*/
+	{"\033[4~",	smallENDkey},	/* End, VT220: SELECT (*6)	*/
 	{"\033[[A",	F1},	/* F1	*/
 	{"\033[[B",	F2},	/* F2	*/
 	{"\033[[C",	F3},	/* F3	*/
 	{"\033[[D",	F4},	/* F4	*/
 	{"\033[[E",	F5},	/* F5	*/
 	{"\033[G",	HOP},	/* 5	*/
-	{"\033[1G",	HOP},	/* pattern for shifted 5? (or some other terminal)	*/
+
+	/* found somewhere? */
+	{"\033[1G",	HOP},	/* pattern for shifted 5? */
 
 	/* PuTTY */
 	{"\033OG",	HOP},	/* 5	*/
@@ -292,7 +499,7 @@ struct fkeyentry fkeymap [] =
 	/* non-application mode xterm */
 	{"\033[H",	smallHOMEkey},	/* Pos1, esp on left keypad	*/
 	{"\033[F",	smallENDkey},	/* End, esp on left keypad	*/
-	{"\033[E",	HOP},	/* 5, Begin	*/
+	{"\033[E",	HOP},	/* 5, Begin, hanterm keypad 5	*/
 
 	/* xterm generic patterns for modified keypad keys */
 	{"\033[1H",	HOMEkey},	/* Pos1	*/
@@ -369,27 +576,53 @@ struct fkeyentry fkeymap [] =
 	{"\033[21~",	F10},	/* F10	*/
 
 	/* VT220-style Fn function keys (control-Fn with xterm VT220 mode) */
-	{"\033[23~",	F11},	/* F11	*/
-	{"\033[24~",	F12},	/* F12	*/
-	/* the following are also attached to Control-Fn by xterm 216 
-	   with the "old function keys" option - it should better 
-	   not be used, especially as the DO (vt220) sequence 
-	   is ambiguously emitted by both Shift-F6 / Control-F6 and 
-	   Menu keys */
+	{"\033[23~",	F11},	/* F11 (also shift-F1)	*/
+	{"\033[24~",	F12},	/* F12 (also shift-F2)	*/
+	/* the following 10 (F3...F12) are also attached to Control-Fn 
+	   by xterm 216 with the "old function keys" option -
+	   it should better not be used, especially as the DO (vt220) 
+	   sequence is ambiguously emitted by both Shift-F6 / Control-F6 
+	   and Menu keys
+	 */
 	{"\033[25~",	F3, shift_mask},	/* F13	*/
 	{"\033[26~",	F4, shift_mask},	/* F14	*/
-	{"\033[28~",	F5, shift_mask},	/* HELP (vt220) */
-	{"\033[29~",	HOP},			/* DO (vt220), Menu (xterm) */
+	{"\033[28~",	F5, shift_mask},	/* VT220: HELP (%1) */
+	{"\033[29~",	HOP},			/* VT220: DO (Redo %0), xterm: Menu */
 	{"\033[31~",	F7, shift_mask},	/* F17	*/
 	{"\033[32~",	F8, shift_mask},	/* F18	*/
 	{"\033[33~",	F9, shift_mask},	/* F19	*/
 	{"\033[34~",	F10, shift_mask},	/* F20	*/
-
 	/* xterm VT220 mode Fn function keys (control-Fn, actually),
-	   with xterm 216, also with oldFunctionKeys mode
+	   since xterm 216, also with oldFunctionKeys mode,
+	   also extended Fn function keys with modifyFunctionKeys disabled
 	 */
-	{"\033[42~",	F11, shift_mask}, /* control-F11 */
-	{"\033[43~",	F12, shift_mask}, /* control-F12 */
+	{"\033[42~",	F11, shift_mask}, /* F11 (also control-F1) */
+	{"\033[43~",	F12, shift_mask}, /* F12 (also control-F2) */
+
+	/* xterm extended Fn function keys with modifyFunctionKeys disabled,
+	   since xterm 216
+	 */
+	{"\033[44~",	F3, ctrl_mask}, /* control-F3 */
+	{"\033[45~",	F4, ctrl_mask}, /* control-F4 */
+	{"\033[46~",	F5, ctrl_mask}, /* control-F5 */
+	{"\033[47~",	F6, ctrl_mask}, /* control-F6 */
+	{"\033[48~",	F7, ctrl_mask}, /* control-F7 */
+	{"\033[49~",	F8, ctrl_mask}, /* control-F8 */
+	{"\033[50~",	F9, ctrl_mask}, /* control-F9 */
+	{"\033[51~",	F10, ctrl_mask}, /* control-F10 */
+	{"\033[52~",	F11, ctrl_mask}, /* control-F11 (also ctrl-shft-F1) */
+	{"\033[53~",	F12, ctrl_mask}, /* control-F12 (also ctrl-shft-F2) */
+
+	{"\033[54~",	F3, ctrlshift_mask}, /* control-shift-F3 */
+	{"\033[55~",	F4, ctrlshift_mask}, /* control-shift-F4 */
+	{"\033[56~",	F5, ctrlshift_mask}, /* control-shift-F5 */
+	{"\033[57~",	F6, ctrlshift_mask}, /* control-shift-F6 */
+	{"\033[58~",	F7, ctrlshift_mask}, /* control-shift-F7 */
+	{"\033[59~",	F8, ctrlshift_mask}, /* control-shift-F8 */
+	{"\033[60~",	F9, ctrlshift_mask}, /* control-shift-F9 */
+	{"\033[61~",	F10, ctrlshift_mask}, /* control-shift-F10 */
+	{"\033[62~",	F11, ctrlshift_mask}, /* control-shift-F11 */
+	{"\033[63~",	F12, ctrlshift_mask}, /* control-shift-F12 */
 
 	/* further VT100 keys - some have been disabled by linux sequences */
 	{"\033[6~",	PD},	/* Next Screen, xterm PgDn	*/
@@ -398,6 +631,7 @@ struct fkeyentry fkeymap [] =
 	{"\033[2~",	INSkey},	/* Insert Here, xterm Ins	*/
 	{"\033[@",	HOMEkey},	/* Pos1, PC-xterm	*/
 	{"\033[e",	ENDkey},	/* End, PC-xterm	*/
+
 	/* application cursor keys */
 	{"\033OA",	MUP},	/* up	*/
 	{"\033OB",	MDN},	/* dn	*/
@@ -405,27 +639,28 @@ struct fkeyentry fkeymap [] =
 	{"\033OD",	MLF},	/* lf	*/
 	{"\033O@",	HOMEkey},	/* Pos1, PC-xterm	*/
 	{"\033Oe",	ENDkey},	/* End, PC-xterm	*/
+
 	/* application keypad */
-	{"\033Op",	INSkey},	/* Ins, keypad 0	*/
-	{"\033Oq",	ENDkey},	/* End, keypad 1	*/
-	{"\033Or",	MDN},	/* dn, keypad 2	*/
-	{"\033Os",	PD},	/* PgDn, keypad 3	*/
-	{"\033Ot",	MLF},	/* lf, keypad 4	*/
-	{"\033Ou",	HOP},	/* mid, keypad 5	*/
-	{"\033Ov",	MRT},	/* rt, keypad 6	*/
-	{"\033Ow",	HOMEkey},	/* Pos1, keypad 7	*/
-	{"\033Ox",	MUP},	/* up, keypad 8	*/
-	{"\033Oy",	PU},	/* PgUp, keypad 9	*/
-	{"\033OM",	SNL},	/* Enter */
-	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
-	{"\033Om",	SU},	/* PF3, keypad - 	*/
-	{"\033On",	DELkey},	/* Del, keypad .	*/
-	/* xterm/vt100 application keypad */
-	{"\033Oo",	I},	/* PF1, keypad /	*/
+	{"\033O ",	HOP},	/* Space, also old hanterm keypad 5 ?	*/
+	{"\033OI",	STAB},	/* shift-Tab */
+	{"\033OM",	SNL},	/* Enter (@8) */
+	{"\033Oo",	FIND},	/* PF1, keypad /	*/
 	{"\033Oj",	I},	/* PF2, keypad *	*/
+	{"\033Om",	SU},	/* PF3, keypad -	*/
 	{"\033Ok",	SD},	/* PF4, keypad +	*/
-	{"\033O ",	HOP},	/* Space, also hanterm keypad 5	*/
-	{"\033OI",	STAB},	/* Tab */
+
+	{"\033Op",	INSkey},	/* Ins (kI), keypad 0	*/
+	{"\033Oq",	ENDkey},	/* End (K4), keypad 1	*/
+	{"\033Or",	MDN},	/* dn (kd), keypad 2	*/
+	{"\033Os",	PD},	/* PgDn (K5), keypad 3	*/
+	{"\033Ot",	MLF},	/* lf (kl), keypad 4	*/
+	{"\033Ou",	HOP},	/* mid (K2), keypad 5	*/
+	{"\033Ov",	MRT},	/* rt (kr), keypad 6	*/
+	{"\033Ow",	HOMEkey},	/* Pos1 (K1), keypad 7	*/
+	{"\033Ox",	MUP},	/* up (ku), keypad 8	*/
+	{"\033Oy",	PU},	/* PgUp (K3), keypad 9	*/
+	{"\033On",	DELkey},	/* Del (kD), keypad .	*/
+	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
 
 	/* Sun */
 	{"\033[208z",	F6, shift_mask},	/* R1 -> grave */
@@ -527,6 +762,13 @@ struct fkeyentry fkeymap [] =
 	/* SCO ansi */
 	{"\033[L",	INSkey},	/* (termcap: insert character) */
 
+	/* xterm mouse */
+	{"\033[M",	DIRECTxterm},	/* direct cursor address */
+	{"\033[t",	TRACKxterm},	/* cursor hilite tracking */
+	{"\033[T",	TRACKxtermT},	/* cursor hilite tracking */
+	{"\033[O",	FOCUSout},	/* window loses focus */
+	{"\033[I",	FOCUSin},	/* window gains focus */
+
 	{NIL_PTR}
 };
 
@@ -663,8 +905,40 @@ struct fkeyentry fkeymap_rxvt [] = {
 	{"\033[3~",	DELkey, ctrl_mask},	/* small keypad Del key	*/
 	/* xterm */
 	{"\033[M",	DIRECTxterm},	/* direct cursor address */
-	{"\033[t",	TRACKxterm},	/* cursor hilite tracking */
-	{"\033[T",	TRACKxtermT},	/* cursor hilite tracking */
+	/* Shift-function keys: shifted by 2 against older conventions */
+	{"\033[23~",	F11},	/* F11, shift-F1 */
+	{"\033[24~",	F12},	/* F12, shift-F2 */
+	{"\033[25~",	F3, shift_mask},
+	{"\033[26~",	F4, shift_mask},
+	{"\033[28~",	F5, shift_mask},
+	{"\033[29~",	F6, shift_mask},
+	{"\033[31~",	F7, shift_mask},
+	{"\033[32~",	F8, shift_mask},
+	{"\033[33~",	F9, shift_mask},
+	{"\033[34~",	F10, shift_mask},
+	{"\033[23$",	F11, shift_mask},
+	{"\033[24$",	F12, shift_mask},
+
+	{NIL_PTR}
+};
+
+struct fkeyentry fkeymap_linux [] = {
+	/* Linux console F-keys */
+	{"\033[23~",	F11},
+	{"\033[24~",	F12},
+	{"\033[25~",	F1, shift_mask},
+	{"\033[26~",	F2, shift_mask},
+	{"\033[28~",	F3, shift_mask},
+	{"\033[29~",	F4, shift_mask},
+	{"\033[31~",	F5, shift_mask},
+	{"\033[32~",	F6, shift_mask},
+	{"\033[33~",	F7, shift_mask},
+	{"\033[34~",	F8, shift_mask},
+
+	/* override terminfo-derived settings */
+	{"\033[1~",	smallHOMEkey},	/* small Pos1	*/
+	{"\033[4~",	smallENDkey},	/* small End	*/
+	{"\033[3~",	smallDELkey},	/* small Del	*/
 
 	{NIL_PTR}
 };
@@ -674,19 +948,33 @@ struct fkeyentry fkeymap_xterm [] = {
 	{"\033[M",	DIRECTxterm},	/* direct cursor address */
 	{"\033[t",	TRACKxterm},	/* cursor hilite tracking */
 	{"\033[T",	TRACKxtermT},	/* cursor hilite tracking */
+	{"\033[O",	FOCUSout},	/* window loses focus */
+	{"\033[I",	FOCUSin},	/* window gains focus */
+
+	/* override wrong terminfo-derived settings */
+	{"\033[1~",	smallHOMEkey},	/* small Pos1	*/
+	{"\033[4~",	smallENDkey},	/* small End	*/
 
 	{NIL_PTR}
 };
 
 struct fkeyentry fkeymap_vt100 [] = {
-	/* VT100 */
-	{"\033[1~",	FIND},	/* Find */
-	{"\033[4~",	MARK},	/* Select */
-	{"\033[29~",	COPY},	/* Do */
-	{"\033[28~",	HELP},	/* Help */
-
 	{"\033OX",	HOP},	/* appl. mode =, conflict with older xterm F9 */
 
+	/* override wrong terminfo-derived settings */
+	{"\033Op",	INSkey},	/* Ins (kI), keypad 0	*/
+	{"\033Oq",	ENDkey},	/* End (K4), keypad 1	*/
+	{"\033Or",	MDN},	/* dn (kd), keypad 2	*/
+	{"\033Os",	PD},	/* PgDn (K5), keypad 3	*/
+	{"\033Ot",	MLF},	/* lf (kl), keypad 4	*/
+	{"\033Ou",	HOP},	/* mid (K2), keypad 5	*/
+	{"\033Ov",	MRT},	/* rt (kr), keypad 6	*/
+	{"\033Ow",	HOMEkey},	/* Pos1 (K1), keypad 7	*/
+	{"\033Ox",	MUP},	/* up (ku), keypad 8	*/
+	{"\033Oy",	PU},	/* PgUp (K3), keypad 9	*/
+	{"\033On",	DELkey},	/* Del (kD), keypad .	*/
+	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
+
 	{NIL_PTR}
 };
 
@@ -862,16 +1150,16 @@ struct fkeyentry fkeymap_siemens [] = {
 	{"\033=",	F10, shift_mask},	/* shift-F10 */
 	{"\033*",	F11, shift_mask},	/* shift-F11 */
 	{"\033+",	F12, shift_mask},	/* shift-F12 */
-	{"\033,",	F3, ctrlshift_mask},	/* shift-F13 */
-	{"\033-",	F4, ctrlshift_mask},	/* shift-F14 */
-	{"\033.",	F5, ctrlshift_mask},	/* shift-F15 */
-	{"\033/",	F6, ctrlshift_mask},	/* shift-F16 */
-	{"\0331",	F7, ctrlshift_mask},	/* shift-F17 */
-	{"\0332",	F8, ctrlshift_mask},	/* shift-F18 */
-	{"\0333",	F9, ctrlshift_mask},	/* shift-F19 */
-	{"\033U",	F10, ctrlshift_mask},	/* shift-F20 */
-	{"\033W",	F11, ctrlshift_mask},	/* shift-F21 */
-	{"\033Y",	F12, ctrlshift_mask},	/* shift-F22 */
+	{"\033,",	F3, ctrlshift_mask},	/* control-shift-F13 */
+	{"\033-",	F4, ctrlshift_mask},	/* control-shift-F14 */
+	{"\033.",	F5, ctrlshift_mask},	/* control-shift-F15 */
+	{"\033/",	F6, ctrlshift_mask},	/* control-shift-F16 */
+	{"\0331",	F7, ctrlshift_mask},	/* control-shift-F17 */
+	{"\0332",	F8, ctrlshift_mask},	/* control-shift-F18 */
+	{"\0333",	F9, ctrlshift_mask},	/* control-shift-F19 */
+	{"\033U",	F10, ctrlshift_mask},	/* control-shift-F20 */
+	{"\033W",	F11, ctrlshift_mask},	/* control-shift-F21 */
+	{"\033Y",	F12, ctrlshift_mask},	/* control-shift-F22 */
 
 	{NIL_PTR}
 };
