@@ -2,26 +2,9 @@
 	mined text editor main include file
 \***************************************************************************/
 
-#ifdef __STDC__
-#define DEFPROTO
-#endif
+#include "_proto.h"
 
-#ifdef __TURBOC__
-#define DEFPROTO
-#endif
-
-
-#ifdef NOPROTO
-#undef DEFPROTO
-#endif
-
-#undef _
-
-#ifdef DEFPROTO
-#define _(x) x
-#else
-#define _(x) ()
-#endif
+#define _FLAG_H
 
 
 /***************************************************************************\
@@ -192,10 +175,11 @@ extern unsigned int usleep ();
 #endif
 
 
+/* this is again unused :) since it broke compilation on SunOS 5.10 */
 #ifdef __GNUC__
-#define unused __attribute__((unused))
+#define __unused__ __attribute__((unused))
 #else
-#define unused
+#define __unused__
 #endif
 
 
@@ -222,7 +206,6 @@ extern int XMAX;
 extern short MENU;
 #define SCREENMAX	(YMAX - 1)	/* last line displayed (first is 0) */
 #define XBREAK		(XMAX - scrollbar_width)	/* Shift line display at this column */
-extern int scrollbar_width;
 #define XBREAK_bottom	(XMAX - 1)
 #define SHIFT_SIZE	(tab8 (XMAX / 4 + 1))	/* Number of chars to shift */
 #define maxLINE_LEN	(maxXMAX + 1)	/* max screen line length */
@@ -230,7 +213,7 @@ extern int scrollbar_width;
 #define MAX_CHARS	1024		/* max chars on one line of text */
 		/* LINE_START must be rounded up to the lowest SHIFT_SIZE */
 #define LINE_START	(((-MAX_CHARS - 1) / SHIFT_SIZE) * SHIFT_SIZE \
-  				   - SHIFT_SIZE)
+				   - SHIFT_SIZE)
 #define LINE_END	(MAX_CHARS * 8)	/* Must be larger than highest 
 					x-coordinate for line; MAX_CHARS + 1
 					is not sufficient due to TABs */
@@ -244,7 +227,7 @@ extern int scrollbar_width;
 #define NO_INPUT	(FINE + 1)
 
 #define STD_IN		0		/* Input file # */
-#define STD_OUT	 	1		/* Terminal output file # */
+#define STD_OUT		1		/* Terminal output file # */
 #define STD_ERR		2
 
 #define REPORT_CHANGED_LINES	1	/* Report change of lines on # lines */
@@ -260,13 +243,11 @@ extern int scrollbar_width;
  */
 typedef enum {
 	releasebutton,
-	leftbutton,
-	middlebutton,
-	rightbutton,
+	leftbutton, middlebutton, rightbutton,
 	movebutton,
-	wheelup,
-	wheeldown
-} mousebutton;
+	wheelup, wheeldown,
+	focusout, focusin
+} mousebutton_type;
 
 #define shift_button	4
 #define meta_button	8
@@ -284,11 +265,11 @@ typedef enum {
   NOT_VALID,
   VALID,
 
-/* Expression flags */
+/* expression flags */
   FORWARD,
   REVERSE,
 
-/* Yank flags */
+/* yank flags */
   SMALLER,
   BIGGER,
   SAME,
@@ -323,9 +304,10 @@ struct Line {
   struct Line * next;
   struct Line * prev;
   char * text;
+  int highlighting_begin, highlighting_end;
   unsigned short shift_count;
   lineend_type return_type;
-  char syntax_marker;	/* bitmask composed of values below */
+  char syntax_marker;	/* bitmask composed of syntax_ flags below */
 };
 #define syntax_none	0
 #define syntax_HTML	1
@@ -401,9 +383,16 @@ extern int input_fd;		/* File descriptors for terminal dialog */
 extern int output_fd;
 
 extern char file_name [];	/* name of file being edited */
-extern PROT fprot0;	/* default prot. mode for new files */
-extern PROT fprot1;	/* prot. mode for new file being edited */
-extern PROT bufprot;	/* prot. mode for paste buffer file */
+extern FLAG writable;		/* Set if file cannot be written */
+extern FLAG file_is_dir;	/* Tried to open a directory as file? */
+extern FLAG reading_pipe;	/* Set if file should be read from stdin */
+extern FLAG writing_pipe;	/* Set if file should be written to stdout */
+extern lineend_type default_lineend;
+
+extern PROT xprot;		/* actual prot. mask representing +x option */
+extern PROT fprot0;		/* default prot. mode for new files */
+extern PROT fprot1;		/* prot. mode for new file being edited */
+extern PROT bufprot;		/* prot. mode for paste buffer file */
 
 #ifdef unix
 extern char * window_string_code;
@@ -429,42 +418,37 @@ extern FLAG dim_HTML;		/* display HTML dimmed ? */
 extern FLAG insert_mode;	/* insert or overwrite */
 extern FLAG append_flag;	/* Set when buffer should be appended to */
 extern FLAG pastebuf_utf8;	/* Paste buffer always treated as UTF-8? */
+extern FLAG rectangular_paste_flag;	/* Copy/Paste rectangular area? */
 extern character erase_char;	/* effective (configured) char for erase left */
 extern char emulation;		/* 'w' for WordStar, 'e' for emacs */
 extern FLAG emacs_buffer;	/* enable emacs buffer fct for ^K/^T */
 extern FLAG paste_stay_left;	/* cursor stays before pasted region */
+extern FLAG tab_left;		/* should moving up/down on TAB go left? */
+extern FLAG tab_right;		/* should moving up/down on TAB go right? */
 extern character quit_char;	/* ^\/^G character to cancel command */
 extern FLAG Turkish;		/* use Turkish case toggle specials ? */
 extern FLAG Lithuanian;		/* use Lithuanian case toggle specials ? */
 extern FLAG smart_quotes;	/* replace " with typographic quote ? */
 
-extern FLAG utf8_screen;	/* screen driven in UTF-8 mode ? */
-extern FLAG utf8_input;		/* keyboard input in UTF-8 mode ? */
-extern FLAG mapped_term;	/* terminal in mapped 8-bit mode ? */
-extern FLAG cjk_term;		/* terminal in CJK mode ? */
-extern FLAG cjk_uni_term;	/* terminal in CJK mode with Unicode widths ? */
-extern FLAG gb18030_term;	/* does CJK terminal support GB18030 ? */
-extern FLAG euc3_term;		/* does CJK terminal support EUC 3 byte ? */
-extern FLAG euc4_term;		/* does CJK terminal support EUC 4 byte ? */
-extern FLAG cjklow_term;	/* does CJK terminal support 81-9F range ? */
-extern int cjk_tab_width;	/* width of CJK TAB indicator */
-extern int cjk_lineend_width;	/* width of CJK line end indicator */
-extern FLAG combining_screen;	/* UTF-8 combining character terminal ? */
-extern FLAG bidi_screen;	/* UTF-8 bidi terminal ? */
-extern int width_data_version;
-#define cjk_width_data 64
-extern int nonbmp_width_data;
-extern int combining_data_version;
-extern FLAG not_assigned_single_width;
 extern FLAG suppress_unknown_cjk;	/* on CJK terminal if no Unicode mapping */
 extern FLAG suppress_extended_cjk;	/* on CJK terminal if in extended code range */
 extern FLAG suppress_invalid_cjk;	/* on CJK terminal if invalid CJK code */
+extern FLAG suppress_surrogates;	/* suppress display of single Unicode surrogates */
+extern FLAG suppress_non_BMP;		/* suppress display of non-BMP range */
+extern FLAG suppress_EF;		/* suppress display of 0x*FFFE/F codes */
+extern FLAG suppress_non_Unicode;	/* suppress display of non-Unicode codes */
+
 extern FLAG utf_cjk_wide_padding;	/* always display CJK on UTF double-width ? */
+
 extern FLAG dark_term;		/* dark colour terminal ? */
+extern FLAG bw_term;		/* black/white terminal ? */
 
-extern int xterm_version;
 extern FLAG configure_xterm_keyboard;	/* deleteIsDEL, metaSendsEscape */
 
+/* from mined1.c */
+extern FLAG auto_detect;	/* Auto detect character encoding from file ? */
+extern char * detect_encodings;	/* List of encodings to detect */
+
 extern FLAG cjk_text;		/* text in CJK encoding ? */
 extern FLAG utf8_text;		/* text in UTF-8 representation ? */
 extern FLAG utf16_file;		/* file encoded in UTF-16 ? */
@@ -472,21 +456,26 @@ extern FLAG utf16_little_endian;	/* UTF-16 file encoded little endian ? */
 extern FLAG mapped_text;	/* text in 8 bit, non-Latin-1 representation ? */
 extern FLAG utf8_lineends;	/* support UTF-8 LS and PS line ends ? */
 extern FLAG poormansbidi;	/* poor man's bidirectional support ? */
+
+extern FLAG combining_mode;	/* UTF-8 combining character support ? */
+extern FLAG separate_isolated_combinings;	/* separated display of comb. at line beg./after TAB ? */
+
 extern FLAG disp_scrollbar;	/* shall scrollbar be displayed ? */
 extern FLAG fine_scrollbar;	/* fine-grained UTF-8 scrollbar ? */
+extern FLAG old_scrollbar_clickmode;	/* old left/right click semantics ? */
+extern int scrollbar_width;
 extern FLAG update_scrollbar_lazy;	/* partial scrollbar refresh as needed ? */
-extern FLAG combining_mode;	/* UTF-8 combining character support ? */
 extern FLAG no_window_title;	/* suppress filename display in window title? */
 
-extern long last_delta_readchar;	/* delay between last 2 characters */
-extern long average_delta_readchar;	/* average delay between last characters */
 extern FLAG loading;		/* Loading a file? Init True for error handling */
+extern FLAG only_detect_text_encoding;
 extern FLAG wordnonblank;	/* handle all non-blank sequences as words */
 extern FLAG proportional;	/* Enable support for proportional fonts? */
+extern int hide_password_mode;	/* Hide / Hide in dot files / Show */
 extern FLAG hide_password;	/* Hide passwords in display */
 extern FLAG translate_output;	/* Transform output diacritics to strings */
-extern char * transout;		/* Output transformation table */
 extern int translen;		/* length of " */
+extern char * transout;		/* Output transformation table */
 extern FLAG controlQS;		/* must respect ^Q/^S handshake ? */
 extern int display_delay;	/* delay between display lines */
 extern FLAG paradisp;		/* Shall paragraph end be distinguished? */
@@ -498,26 +487,28 @@ extern char TAB0_marker;	/* Char to be shown at start of tab chars */
 extern char TAB2_marker;	/* Char to be shown at end of tab chars */
 extern char TABmid_marker;	/* Char to be shown in middle of tab chars */
 extern unsigned long CJK_TAB_marker;	/* Char to be shown in place of tab */
-extern char RET_marker;		/* Char indicating end of line */
-extern char DOSRET_marker;	/* Char indicating DOS end of line */
+extern char RET_marker;		/* Char indicating end of line (LF) */
+extern char DOSRET_marker;	/* Char indicating DOS end of line (CRLF) */
+extern char MACRET_marker;	/* Char indicating Mac end of line (CR) */
 extern char PARA_marker;	/* Char indicating end of paragraph */
 extern char RETfill_marker;	/* Char to fill the end of line with */
 extern char RETfini_marker;	/* Char to fill last position of line with */
 extern char SHIFT_marker;	/* Char indicating that line continues */
 extern char SHIFT_BEG_marker;	/* Char indicating that line continues left */
+extern char MENU_marker;	/* Char to mark selected item */
 extern char * UTF_TAB_marker;		/* Char to be shown in place of tab chars */
 extern char * UTF_TAB0_marker;		/* Char to be shown at start of tab chars */
 extern char * UTF_TAB2_marker;		/* Char to be shown at end of tab chars */
 extern char * UTF_TABmid_marker;	/* Char to be shown in middle of tab chars */
 extern char * UTF_RET_marker;		/* Char indicating end of line */
 extern char * UTF_DOSRET_marker;	/* Char indicating DOS end of line */
+extern char * UTF_MACRET_marker;	/* Char indicating Mac end of line */
 extern char * UTF_PARA_marker;		/* Char indicating end of paragraph */
 extern char * UTF_RETfill_marker;	/* Char to fill the end of line with */
 extern char * UTF_RETfini_marker;	/* Char to fill last position of line with */
 extern char * UTF_SHIFT_marker;		/* Char indicating that line continues */
 extern char * UTF_SHIFT_BEG_marker;	/* Char indicating that line continues left */
-
-extern unsigned long menu_marker;	/* ✓ */
+extern char * UTF_MENU_marker;	/* Char to mark selected item */
 
 extern FLAG stat_visible;	/* Set if status line is visible */
 extern FLAG top_line_scrolled;	/* Was menu line scrolled away? */
@@ -525,6 +516,15 @@ extern FLAG top_line_scrolled;	/* Was menu line scrolled away? */
 extern FLAG page_scroll;	/* use scroll for page up/down */
 extern FLAG page_stay;		/* stay at edge of page after page up/down */
 
+/* from justify.c */
+extern int first_left_margin;
+extern int next_left_margin;
+extern int right_margin;
+
+/* from keyboard.c */
+extern long last_delta_readchar;	/* delay between last 2 characters */
+extern long average_delta_readchar;	/* average delay between last characters */
+
 
 /***************************************************************************\
 	declaration of mined functions
@@ -539,21 +539,15 @@ extern void move_to _((int x, int y));
 extern void move_address _((char *, int y));
 extern void move_address_w_o_RD _((char *, int y));
 extern int find_x _((LINE * line, char * address));
+extern int get_cur_pos _((void));
 extern int get_cur_col _((void));
 /* from edit.c */
 extern int find_y _((LINE * line));
 extern int find_y_w_o_RD _((LINE * line));
 
-/* file handling */
-/* from textfile.c */
-extern int get_line _((int fd, char * buffer, int * len));
-extern int panicwrite _((void));
-
 /* text buffer status handling */
 /* from mined1.c */
 extern void set_modified _((void));
-/* from textfile.c */
-extern FLAG checkoverwrite _((char *));
 /* from textbuf.c */
 extern void file_status _((char * message, long bytecount, long charcount, 
 			char * filename, int lines, FLAG textstat, 
@@ -576,20 +570,13 @@ extern void delete_text_buf _((LINE * start_line, char * start_textp, LINE * end
 /* editing functions */
 /* from edit.c */
 extern int delete_text _((LINE * startl, char * start, LINE * endl, char * end));
+extern int insert_text _((LINE *, char * location, char * string));
 extern LINE * line_insert _((LINE *, char *, int, lineend_type));
-extern int insert _((LINE *, char * location, char * string));
 
 /* from pastebuf.c */
 extern void yank_HTML _((FLAG remove));
 extern void paste_HTML _((void));
 
-extern lineend_type default_lineend;
-/* from textfile.c */
-extern void edit_nth_file _((int));
-extern FLAG save_text_load_file _((char * filename));
-extern lineend_type extract_lineend_type _((char *, int));
-
-
 /* character input handling */
 struct prefixspec {
 	voidfunc prefunc;
@@ -602,6 +589,7 @@ struct prefixspec {
 };
 
 /* from compose.c */
+extern char * mnemos _((int ucs));
 extern unsigned long compose _((unsigned long, unsigned long));
 extern unsigned long compose_mnemonic _((char *));
 extern struct prefixspec * lookup_prefix _((voidfunc prefunc, unsigned int shift));
@@ -645,16 +633,12 @@ extern FLAG expand_tabs;	/* Expand TABs to Spaces? */
 #define alpha(c)	((c) != ' ' && (c) != '\t' && (c) != '\n')
 
 
-/* from textfile.c */
-extern void reset_get_line _((FLAG from_text_file /* consider UTF-16 ? */));
-extern void show_get_l_errors _((void));
-
 /* from edit.c */
 extern int get_idf _((char * idf_buf, char * text, char * start_line));
 
 /* from search.c */
-extern void search_for _((char *, FLAG));
-extern void search_expr _((char *, FLAG));
+extern void search_for _((char *, FLAG direction, FLAG case_ignore));
+extern void search_expr _((char *, FLAG direction, FLAG case_ignore));
 extern void search_corresponding _((char *, FLAG, char *));
 
 /* from mined1.c */
@@ -662,11 +646,6 @@ extern void viewonlyerr _((void));
 extern void restrictederr _((void));
 
 
-/* file output functions */
-/* from textfile.c */
-extern int writechar _((int fd, char));
-extern int write_lineend _((int fd, lineend_type return_type, FLAG handle_utf16));
-
 /* special stuff */
 /* from minedaux.c */
 extern char * envvar _((char * name));
@@ -681,6 +660,7 @@ extern char * serrorof _((int));
 extern int geterrno _((void));
 extern void delete_file _((char *));
 extern char * unnull _((char *));
+extern char * num_out _((long number, long radix));
 extern void copy_string _((char * to, char * from));
 
 /* memory allocation */
@@ -705,22 +685,18 @@ extern long get_number _((char * message, char firstdigit, int * result));
 extern int prompt_num_char _((unsigned long * result, unsigned long maxvalue));
 
 
-/* character properties */
-struct scriptentry {
-	unsigned long first, last;
-	char * scriptname;
-	char * categoryname;
-};
 /* from edit.c */
-extern struct scriptentry * scriptinfo _((unsigned long ucs));
 extern char * script _((unsigned long ucs));
+extern unsigned long decomposition_base _((unsigned long ucs));
 extern char * category _((unsigned long ucs));
 extern FLAG isLetter _((unsigned long ucs));
+extern FLAG is_wideunichar _((unsigned long ucs));
 extern FLAG iscombining_unichar _((unsigned long ucs));
+extern FLAG isspacingcombining_unichar _((unsigned long ucs));
 extern FLAG iscombined_unichar _((unsigned long ucs, char * charpos, char * linebegin));
 extern unsigned long case_convert _((unsigned long unichar, int dir));
 /* from charcode.c */
-extern int iscontrol _((character));
+extern int iscontrol _((unsigned long));
 extern character controlchar _((character));
 
 /* scrollbar handling */
@@ -728,6 +704,7 @@ extern character controlchar _((character));
 extern FLAG display_scrollbar _((FLAG update));
 extern void scrollbar_scroll_up _((int from));
 extern void scrollbar_scroll_down _((int from));
+extern int check_scrollbar_pos _((void));
 
 /* output functions */
 extern void clear_filebuf _((void));
@@ -745,12 +722,12 @@ extern FLAG marker_defined _((character marker, char * utfmarker));
 extern void put_blanks _((int endpos));
 extern void set_cursor_xy _((void));
 extern void put_line _((int scr_y, LINE *, int offset, FLAG clear, FLAG prop_pos));
+extern void print_line _((int scr_y, LINE *));
 /* put character values */
 #define putchar(c)	__putchar ((character) c)
 extern void putcharacter _((character));
 extern void put_unichar _((unsigned long));
 extern int put_cjkchar _((unsigned long));
-extern int put_cjktermchar _((unsigned long));
 
 /* character handling */
 /* from minedaux.c */
@@ -777,17 +754,11 @@ extern void utf8_info _((char *, int *, unsigned long *));
 extern int isjoined _((unsigned long, char *, char *));
 extern int iscombined _((unsigned long, char *, char *));
 #define multichar(c)	((character) c >= 0x80 && (cjk_text == False || text_encoding_tag != 'S' || (character) c < 0xA1 || (character) c > 0xDF))
-extern int iscombining _((unsigned long));
+extern int iscombining _((unsigned long ucs));
+extern int iswide _((unsigned long ucs));
 
-/* from iscombin.c, iswide.c, isrtol.c */
-extern int iscombining_char _((unsigned long));
-extern int iswide _((unsigned long));
+/* from isrtol.c */
 extern int is_right_to_left _((unsigned long ucs));
-/* from encoding.c */
-extern char * locale_terminal_encoding _((void));
-extern char * locale_text_encoding _((void));
-extern char * language_code;
-extern char * country_code;
 
 /* from charcode.c */
 extern int utfencode _((unsigned long, character *));
@@ -822,10 +793,14 @@ extern char text_encoding_tag;
 extern char * text_encoding_flag;
 extern char term_encoding_tag;
 
-/* character description flags */
+/* information display preferences */
 extern FLAG always_disp_fstat;	/* Permanent file status display on status line? */
 extern FLAG always_disp_help;	/* Permanent F2... help display on status line? */
 extern FLAG always_disp_code;	/* Permanent char code display on status line? */
+extern FLAG disp_scriptname;	/* display Unicode script range? */
+extern FLAG disp_charname;	/* display Unicode character name? */
+extern FLAG disp_decomposition;	/* display Unicode character decomposition? */
+extern FLAG disp_mnemos;	/* display mined input mnemos? */
 extern FLAG always_disp_Han;	/* Permanent Han character description display on status line? */
 extern FLAG disp_Han_Mandarin;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Cantonese;	/* display this Han pronunciation ? */
@@ -835,9 +810,10 @@ extern FLAG disp_Han_Hangul;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Korean;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Vietnamese;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_HanyuPinlu;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_XHCHanyuPinyin;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Tang;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_description;	/* display Han description ? */
-extern FLAG disp_Han_full;	/* display full Han description ? */
+extern FLAG disp_Han_full;	/* display full popup Han description ? */
 
 /* Han character descriptions table */
 struct hanentry {
@@ -850,6 +826,7 @@ struct hanentry {
 	char * Korean;
 	char * Vietnamese;
 	char * HanyuPinlu;
+	char * XHCHanyuPinyin;
 	char * Tang;
 	char * Definition;
 };
@@ -901,7 +878,8 @@ typedef struct {
 	char * extratag;
 	} menuitemtype;
 extern void separator _((void));
-extern mousebutton mouse_button, mouse_lastbutton, mouse_prevbutton;
+extern mousebutton_type mouse_button, mouse_lastbutton, mouse_prevbutton;
+extern FLAG window_focus;	/* does the window have mouse/keyboard focus? */
 extern int mouse_shift;
 extern int mouse_xpos, mouse_ypos, mouse_lastxpos, mouse_lastypos;
 extern FLAG report_release;
@@ -943,15 +921,16 @@ extern void invoke_key_function _((unsigned long));
 extern void QUICKMENU _((void)), FILEMENU _((void)), EDITMENU _((void));
 extern void SEARCHMENU _((void)), PARAMENU _((void)), OPTIONSMENU _((void));
 extern void handleFlagmenus _((void));
-extern void MOUSEescape _((void));
+extern void MOUSEescape _((void)), FOCUSout _((void)), FOCUSin _((void));
 extern void ANSIseq _((void));
 extern void MUP _((void)), MDN _((void)), MLF _((void)), MRT _((void));
 extern void MNW _((void)), MPW _((void));
 extern void BSEN _((void)), ESEN _((void));
 extern void SD _((void)), SU _((void)), PD _((void)), PU _((void));
 extern void BFILE _((void)), EFILE _((void));
-extern void BLINE _((void)), ELINE _((void)), HIGH _((void)), LOW _((void));
-extern void MDNEL _((void)), MPPARA _((void)), MNPARA _((void));
+extern void BLINE _((void)), ELINE _((void)), GO_LINEBEGIN _((void)), GO_LINEEND _((void));
+extern void HIGH _((void)), LOW _((void));
+extern void MPPARA _((void)), MNPARA _((void));
 extern void DPC _((void)), DCC _((void)), DLN _((void)), DNW _((void)), DPW _((void));
 extern void DCC0 _((void)), DPC0 _((void));
 extern void CTRLINS _((void)), DLINE _((void)), TOGINS _((void)), ctrlQ _((void)), ctrlK _((void)), ctrlO _((void));
@@ -962,7 +941,7 @@ extern void RD _((void)), RDwin _((void)), RDwin_nomenu _((void)), I _((void));
 extern void EXED _((void)), VIEW _((void));
 extern void HOP _((void)), CANCEL _((void)), ESCAPE _((void));
 extern void EMAX _((void)), META _((void)), UNDO _((void));
-extern void CHDI _((void)), PRINT _((void)), CMD _((void)), SH _((void)), SUSP _((void));
+extern void CHDI _((void)), PRINT _((void)), SH _((void)), SUSP _((void));
 extern void LNCI _((void)), LNSW _((void));
 extern void LOWCAP _((void)), LOWER _((void)), UPPER _((void));
 extern void CAPWORD _((void)), LOWCAPWORD _((void));
@@ -978,6 +957,7 @@ extern void SFW _((void)), SRV _((void)), RS _((void)), RSreverse _((void));
 extern void LR _((void)), GR _((void)), REPL _((void));
 extern void SIDFW _((void)), SIDRV _((void));
 extern void Stag _((void));
+extern void AltX _((void));
 
 extern void toggleKEYMAP _((void)), setupKEYMAP _((void));
 extern void toggle_encoding _((void));
@@ -1001,6 +981,7 @@ extern void WB _((void)), INSFILE _((void));
 
 extern void SIDF _((FLAG)), SCURCHAR _((FLAG));
 extern void UML _((char)), REPT _((char));
+extern FLAG CONV _((void));
 extern void BADch _((unsigned long cmd));
 extern void SCORR _((FLAG pref_direction));
 
@@ -1012,12 +993,14 @@ extern void display_code _((void)), display_the_code _((void));
 extern void display_Han _((char * cpoi, FLAG force_utf8));
 extern void display_FHELP _((void));
 extern void ABOUT _((void));
-extern void HELP _((void)), HELP_topics _((void)), select_FHELP _((void)), toggle_FHELP _((void));
+extern void HELP _((void)), HELP_topics _((void)), HELP_Fn _((void)), HELP_accents _((void)), toggle_FHELP _((void));
 extern void FHELP _((voidfunc));
 
 /* function keys */
 extern void HOMEkey _((void)), ENDkey _((void)), smallHOMEkey _((void)), smallENDkey _((void));
-extern void INSkey _((void)), DELkey _((void)), DELchar _((void));
+extern void INSkey _((void)), DELkey _((void)), smallINSkey _((void)), smallDELkey _((void));
+extern void DELchar _((void));
+extern void FIND _((void)), AGAIN _((void));
 extern void F1 _((void)), F2 _((void)), F3 _((void)), F4 _((void)), F5 _((void)), F6 _((void)), F7 _((void)), F8 _((void)), F9 _((void)), F10 _((void)), F11 _((void)), F12 _((void));
 
 
@@ -1028,11 +1011,6 @@ extern void F1 _((void)), F2 _((void)), F3 _((void)), F4 _((void)), F5 _((void))
 #define strisprefix(s1, s2)	(strncmp (s2, s1, strlen (s1)) == 0)
 #define strcontains(s1, s2)	(strstr (s1, s2))
 
-/*
- * Print line on terminal at offset 0 and clear tail of line
- */
-#define line_print(scry, line)	put_line (scry, line, 0, True, False)
-
 /*
  * Functions handling status_line.
  */
@@ -1041,6 +1019,8 @@ extern void status_uni _((char *));
 #define status_msg(str)		status_line (str, NIL_PTR)
 #define status_line(str1, str2)	\
 	(void) bottom_line (True, (str1), (str2), NIL_PTR, False, "")
+#define status_fmt2(str1, str2)	\
+	(void) bottom_line (VALID, str1, str2, NIL_PTR, False, "");
 #define status_beg(str)		\
 	(void) bottom_line (True, (str), NIL_PTR, NIL_PTR, True, "")
 #define error(str1)		\
