@@ -5,6 +5,7 @@
 
 #include "mined.h"
 #include "io.h"
+#include "termprop.h"
 
 #include <errno.h>
 
@@ -43,6 +44,20 @@ disp_scrollbar_end ()
 static int prev_disp_start = 0;
 static int prev_disp_end = 0;
 
+int
+check_scrollbar_pos ()
+{
+	int unit = utf8_screen && fine_scrollbar ? 8 : 1;
+	int mouse_pos = mouse_ypos * unit;
+	if (mouse_pos < prev_disp_start) {
+		return -1;
+	} else if (mouse_pos > prev_disp_end) {
+		return 1;
+	} else {
+		return 0;
+	}
+}
+
 static char shiftskipdouble = '';	/* indicator for shifted double width char */
 
 static FLAG scrollbar_dirty = True;	/* does scrollbar need refresh ? */
@@ -401,6 +416,20 @@ display_scrollbar (update)
 #endif
 
 
+/*======================================================================*\
+|*			Screen function					*|
+\*======================================================================*/
+
+void
+clearscreen ()
+{
+  clear_screen ();
+  top_line_scrolled = True;
+  set_scrollbar_dirty (0);
+  set_scrollbar_dirty (SCREENMAX - 1);
+}
+
+
 /*======================================================================*\
 |*			Marked character output				*|
 \*======================================================================*/
@@ -477,21 +506,27 @@ putnarrowutf (utfmarker)
 
 
 #define indicate_illegal_UTF indicate_marked
-#define indicate_UTF indicate_marked
+#define indicate_narrow indicate_marked
+#define indicate_marked(c) indicate_uni(c, False)
 
 /**
    Indicate highlighted special character property (e.g. illegal code).
  */
 static
 void
-indicate_marked (c)
-  character c;
+indicate_uni (c, allow_wide)
+  unsigned long c;
+  FLAG allow_wide;
 {
   if (char_on_status_line) {
 	reverse_off ();
   }
   unidisp_on ();
-  putnarrowchar ((unsigned long) c);
+  if (allow_wide) {
+	put_unichar (c);
+  } else {
+	putnarrowchar (c);
+  }
   unidisp_off ();
   if (char_on_status_line) {
 	reverse_on ();
@@ -522,7 +557,7 @@ indicate_special (c)
  * putmarkmode and endmarkmode are only called by putmark and for 
  * display of TAB markers
  */
-static FLAG dim_mode = False;
+static FLAG mark_mode_active = False;
 static FLAG mark_alt_cset = False;
 
 static
@@ -532,9 +567,9 @@ putmarkmode (marker, utfmarker, dim_me)
   char * utfmarker;
   FLAG dim_me;
 {
-  if (dim_me && ! dim_mode) {
-	dim_on ();
-	dim_mode = True;
+  if (dim_me && ! mark_mode_active) {
+	mark_on ();
+	mark_mode_active = True;
   }
 
   if (utf8_screen) {
@@ -562,7 +597,7 @@ putmarkmode (marker, utfmarker, dim_me)
 		altcset_off ();
 		mark_alt_cset = False;
 	}
-	if (mapped_term) {
+	if (mapped_term || cjk_term) {
 		unsigned long termchar = mappedtermchar (marker);
 		if (no_char (termchar)) {
 			__putchar (char7bit (marker));
@@ -579,9 +614,9 @@ static
 void
 endmarkmode ()
 {
-  if (dim_mode) {
-	dim_off ();
-	dim_mode = False;
+  if (mark_mode_active) {
+	mark_off ();
+	mark_mode_active = False;
   }
   if (mark_alt_cset) {
 	altcset_off ();
@@ -601,22 +636,6 @@ putmark (marker, utfmarker)
   endmarkmode ();
 }
 
-static
-void
-putCJKmark (umark)
-  unsigned long umark;
-{
-  unsigned long mark = mappedtermchar (umark);
-  dim_on ();
-  if (no_char (mark)) {
-	put_cjkchar ('?');
-	put_cjkchar ('?');
-  } else {
-	put_cjktermchar (mark);
-  }
-  dim_off ();
-}
-
 static
 void
 putUmark (marker, utfmarker)
@@ -629,15 +648,32 @@ putUmark (marker, utfmarker)
   unimarkdisp_off ();
 }
 
+void
+put_menu_marker ()
+{
+/*
+	unidisp_on ();
+	If utf8_screen
+	Then	put_unichar (menu_marker);
+	Else	putchar ('*');
+	Fi
+	unidisp_off ();
+*/
+  unidisp_on ();
+  putmarkmode (MENU_marker, UTF_MENU_marker, False);
+  endmarkmode ();
+  unidisp_off ();
+}
+
 static
 void
 putshiftmark (marker, utfmarker)
   character marker;
   char * utfmarker;
 {
-  if (! dim_mode) {
-	dim_on ();
-	dim_mode = True;
+  if (! mark_mode_active) {
+	mark_on ();
+	mark_mode_active = True;
   }
   reverse_on ();
   putmarkmode (marker, utfmarker, False);
@@ -653,6 +689,8 @@ marker_defined (marker, utfmarker)
   return marker || (utf8_screen && utfmarker && * utfmarker);
 }
 
+#define token_output
+
 static
 void
 putret (type)
@@ -662,27 +700,98 @@ putret (type)
 	putmark (RET_marker, UTF_RET_marker);
   } else if (type == lineend_CRLF) {
 	/* MSDOS line separator */
+#ifdef token_output
+	set_colour_token (4);	/* blue */
+#endif
 	putmark (DOSRET_marker, UTF_DOSRET_marker);
   } else if (type == lineend_CR) {
 	/* Mac line separator */
-/*	putmark ('', "¥");	*/
-/*	putmark ('', "¦");	*/
-	putmark ('@', "@");
+#ifdef token_output
+	set_colour_token (3);	/* yellow */
+#endif
+	putmark (MACRET_marker, UTF_MACRET_marker);
   } else if (type == lineend_NONE) {
 	putmark ('', "¬");
   } else if (type == lineend_NUL) {
 	putmark ('', "º");
   } else if (type == lineend_LS) {
-	/* Unicode line separator 2028:   */
+	/* Unicode line separator U+2028:   */
 	putUmark (RET_marker, UTF_RET_marker);
   } else if (type == lineend_PS) {
-	/* Unicode paragraph separator 2029:   */
+	/* Unicode paragraph separator U+2029:   */
 	putUmark (PARA_marker, UTF_PARA_marker);
   } else {
 	putmark ('', "¤");
   }
 }
 
+static int put_cjktermchar _((unsigned long cjkchar));
+
+static
+void
+putCJKtab ()
+{
+  unsigned long mark = mappedtermchar (CJK_TAB_marker);
+  mark_on ();
+  if (no_char (mark)) {
+	put_cjkchar ('.');
+	put_cjkchar ('.');
+  } else {
+	put_cjktermchar (mark);
+  }
+  mark_off ();
+}
+
+static
+void
+putCJKret (type)
+  unsigned char type;
+{
+/* markers available in all CJK encodings:
+	300A	《
+	FF20	＠
+	03BC	μ
+	00B0	°
+	00A7	§
+*/
+  unsigned long umark;
+  unsigned long mark;
+
+  if (type == lineend_LF) {
+	umark = 0x300A;	/* 《 */
+  } else if (type == lineend_CRLF) {
+	/* MSDOS line separator */
+	umark = 0x03BC;	/* μ */
+  } else if (type == lineend_CR) {
+	/* Mac line separator */
+	umark = 0xFF20;	/* ＠ */
+  } else if (type == lineend_NONE) {
+	umark = '-';
+  } else if (type == lineend_NUL) {
+	umark = 0x00B0;	/* ° */
+  } else if (type == lineend_LS) {
+	/* Unicode line separator U+2028 */
+	putUmark ('<', NIL_PTR);
+	return;
+  } else if (type == lineend_PS) {
+	/* Unicode paragraph separator U+2029 */
+	putUmark (0x00A7, NIL_PTR);
+	return;
+  } else {
+	umark = 0x300A;	/* 《 */
+  }
+
+  mark = mappedtermchar (umark);
+  mark_on ();
+  if (no_char (mark)) {
+	put_cjkchar ('<');
+	put_cjkchar ('<');
+  } else {
+	put_cjktermchar (mark);
+  }
+  mark_off ();
+}
+
 
 /*======================================================================*\
 |*			Character output				*|
@@ -701,7 +810,7 @@ struct interval {
 struct interval list_Quotation_Mark [] =
 struct interval list_Dash [] =
 */
-#include "charprop.h"
+#include "charprop.t"
 
 static
 int
@@ -738,6 +847,18 @@ lookup (ucs, table, length)
 #define trace_put_cjk(tag, c)	
 #endif
 
+/**
+   check if character is not a valid (assigned) Unicode terminal character
+   (i.e. considering the auto-detected Unicode version of the terminal)
+ */
+static
+FLAG
+no_validunichar (u)
+  unsigned long u;
+{
+  return no_unichar (u) || ! term_isassigned (u);
+}
+
 /**
    put CJK character to screen;
    to be called only if cjk_term == True
@@ -847,7 +968,7 @@ put_cjkcharacter (term, cjkchar, width)
 		}
 		return width;
 	} else if (suppress_unknown_cjk && 
-		   no_unichar (term
+		   no_validunichar (term
 				? lookup_mappedtermchar (cjkchar) 
 				: lookup_encodedchar (cjkchar))
 		  ) {
@@ -910,6 +1031,7 @@ put_cjkchar (cjkchar)
    to be called only if cjk_term == True
    returns screen length
  */
+static
 int
 put_cjktermchar (cjkchar)
   unsigned long cjkchar;
@@ -917,6 +1039,30 @@ put_cjktermchar (cjkchar)
   return put_cjkcharacter (True, cjkchar, -1);
 }
 
+/**
+   put joined character for separate display to screen
+ */
+static
+void
+put_unjoined (unichar)
+  unsigned long unichar;
+{
+  /* prevent terminal ligature by substituting joining character 
+     with its ISOLATED FORM */
+  unsigned long subst = isolated_alef (unichar);
+  if ((cjk_term || mapped_term) && no_char (mappedtermchar (subst))) {
+	/* use TATWEEL instead; could use one of
+		U+061F ؟ ARABIC QUESTION MARK
+		U+066D ٭ ARABIC FIVE POINTED STAR
+		U+0640 ـ ARABIC TATWEEL
+	   (or rather substitute the base char U+0644 ل ARABIC LETTER LAM)
+	 */
+	put_unichar (0x0640);
+  } else {
+	put_unichar (subst);
+  }
+}
+
 
 static
 void
@@ -924,11 +1070,14 @@ put_uniend ()
 {
   if (utf8_text && utfcount > 0) {
 	/* previous character not terminated properly */
-	indicate_illegal_UTF ('');
+	indicate_illegal_UTF ((character) '');
 	utfcount = 0;
   }
 }
 
+
+#define REPLACEMENT_CHARACTER 0xFFFD
+
 /*
  * put Unicode (UCS-4, actually) character to screen;
  * put as UTF-8 on UTF-8 terminal, or as Latin-1 otherwise
@@ -943,24 +1092,49 @@ void
 put_unichar (unichar)
   unsigned long unichar;
 {
+  if (unassigned_single_width) {
+	if (rxvt_version > 0) {
+		/* handle weird mapping of non-Unicode ranges */
+		if (unichar < 0x80000000) {
+			unichar &= 0x1FFFFF;
+		}
+	}
+  }
+
   if (mapped_text && no_unichar (unichar)) {
 	indicate_marked ('?');
   } else if (unichar < ' ') {
-	indicate_UTF (unichar + '@');
+	indicate_narrow (unichar + '@');
   } else if (unichar == 0x7F) {
-	indicate_UTF ('?');
+	indicate_narrow ('?');
   } else if (unichar >= 0x80 && unichar <= 0x9F) {
 	if (unichar == 0x9F) {
-		indicate_UTF ('?');
+		indicate_narrow ('?');
 	} else {
-		indicate_UTF (unichar - 0x20);
+		indicate_narrow (unichar - 0x20);
 	}
   } else if (unichar == 0xA0 /* nbsp */) {
 	indicate_special ('');	/* indicate nbsp in UTF-8 / mapped terminal */
-  } else if (unichar == 0xFFFE) {
-	indicate_illegal_UTF ('e');
-  } else if (unichar == 0xFFFF) {
-	indicate_illegal_UTF ('f');
+  } else if (suppress_EF && (unichar & 0xFFFE) == 0xFFFE) {
+	indicate_illegal_UTF (REPLACEMENT_CHARACTER);
+	if (iswide (unichar)) {
+		indicate_illegal_UTF (' ');
+	}
+  } else if (suppress_non_Unicode && unichar >= 0x110000 && unichar < 0x80000000) {
+	indicate_illegal_UTF (REPLACEMENT_CHARACTER);
+	if (iswide (unichar)) {
+		indicate_illegal_UTF (' ');
+	}
+  } else if (suppress_non_BMP && unichar >= 0x10000) {
+	indicate_illegal_UTF (REPLACEMENT_CHARACTER);
+	if (iswide (unichar)) {
+		indicate_illegal_UTF (' ');
+	}
+  } else if (suppress_surrogates && unichar >= 0xD800 && unichar <= 0xDFFF) {
+	indicate_illegal_UTF (REPLACEMENT_CHARACTER);
+	if (iswide (unichar)) {
+		indicate_illegal_UTF (' ');
+	}
   } else if (utf8_screen) {
 	if (unichar < 0x80) {
 		__putchar (unichar);
@@ -1002,17 +1176,41 @@ put_unichar (unichar)
 			*/
 		}
 	}
-  } else {
+  } else {	/* not utf8_screen */
 	unsigned long termchar;
+	int iswide_unichar;
+	int iswide_base;
+	static char * dummyline = "";
+
+	if (cjk_term || mapped_term) {
+		termchar = mappedtermchar (unichar);
+	}
+
+#ifdef wrong_attempt
+	if (cjk_term || mapped_term) {
+		termchar = mappedtermchar (unichar);
+		iswide_unichar = cjkscrwidth (termchar, dummyline, dummyline) == 2;
+	} else {
+		iswide_unichar = iswide (unichar);
+	}
+#else
+	/* here the width assumed by the caller of this function is needed */
+	iswide_unichar = iswide (unichar);
+#endif
 
 	if (cjk_term) {
-		termchar = mappedtermchar (unichar);
 		if (! no_char (termchar)) {
-			put_cjktermchar (termchar);
+			if (utf8_text && ! cjk_uni_term) {
+				/* workaround for inconsistent width handling 
+				   when replacing Unicode with "#" (-> "# ")
+				 */
+				(void) put_cjkcharacter (True, termchar, -1);
+			} else {
+				(void) put_cjkcharacter (True, termchar, 1 + iswide_unichar);
+			}
 			return;
 		}
 	} else if (mapped_term) {
-		termchar = mappedtermchar (unichar);
 		if (! no_char (termchar)) {
 			__putchar (termchar);
 			return;
@@ -1025,56 +1223,91 @@ put_unichar (unichar)
 	/* character cannot be displayed on terminal;
 	   display indication or substitution */
 	if (iscombining_unichar (unichar)) {
-		indicate_UTF ('\'');
+		indicate_narrow ('\'');
 	} else {
-		if (unichar >= 0xFF01 && unichar <= 0xFF5E) {
-			/* FULLWIDTH forms */
-			indicate_UTF (unichar - 0xFEE0);
-		} else if (unichar == 0xFFE0) {
-			/* FULLWIDTH CENT SIGN */
-			indicate_UTF (0xA2);
-		} else if (unichar == 0xFFE1) {
-			/* FULLWIDTH POUND SIGN */
-			indicate_UTF (0xA3);
-		} else if (unichar == 0xFFE2) {
-			/* FULLWIDTH NOT SIGN */
-			indicate_UTF (0xAC);
-		} else if (unichar == 0xFFE3) {
-			/* FULLWIDTH MACRON */
-			indicate_UTF (0xAF);
-		} else if (unichar == 0xFFE4) {
-			/* FULLWIDTH BROKEN BAR */
-			indicate_UTF (0xA6);
-		} else if (unichar == 0xFFE5) {
-			/* FULLWIDTH YEN SIGN */
-			indicate_UTF (0xA5);
-		} else if (unichar == 0x20AC) {
+		if (unichar == 0x20AC) {
 			/* EURO SIGN */
-			indicate_UTF ('E');
+			indicate_narrow ('E');
+		} else if (unichar == 0x2713) {
+			/* ✓ */
+			indicate_narrow ('V');
 		} else if (lookup (unichar, list_Quotation_Mark, arrlen (list_Quotation_Mark))) {
 			if (unichar == 0xAB || unichar == 0x2039) {
-				indicate_UTF ('<');
+				indicate_narrow ('<');
 			} else if (unichar == 0xBB || unichar == 0x203A) {
-				indicate_UTF ('>');
+				indicate_narrow ('>');
 			} else {
-				indicate_UTF ('"');
+				indicate_narrow ('"');
 			}
 		} else if (lookup (unichar, list_Dash, arrlen (list_Dash))) {
 			if (unichar == 0x301C || unichar == 0x3030) {
-				indicate_UTF ('~');
+				indicate_narrow ('~');
 			} else {
-				indicate_UTF ('-');
+				indicate_narrow ('-');
 			}
-		} else if (unichar != (character) UNI_marker) {
-			indicate_UTF (UNI_marker);
 		} else {
-			indicate_UTF (char7bit (UNI_marker));
+			unsigned long base = unichar;
+			FLAG printable;
+
+#define dont_debug_subst_decompose
+#ifdef debug_subst_decompose
+			printf ("%04X(w=%d)\n", base, 1+iswide (base));
+#endif
+
+			do {
+				printable = True;
+				base = decomposition_base (base);
+				if (cjk_term || mapped_term) {
+					termchar = mappedtermchar (base);
+					if (no_char (termchar)) {
+						printable = False;
+					}
+				} else if (base >= 0x100) {
+					printable = False;
+				}
+#ifdef debug_subst_decompose
+				printf ("      -> %04X(w=%d) %sprintable\n", base, 1+iswide (base), printable ? "" : "not ");
+#endif
+			} while (base > 0 && ! printable);
+
+			if (cjk_term || mapped_term) {
+				iswide_base = cjkscrwidth (termchar, dummyline, dummyline) == 2;
+			} else {
+				iswide_base = iswide (base);
+			}
+#ifdef debug_subst_decompose
+			printf ("    unichar %04X w=%d -> base %04X w=%d\n", 
+				unichar, 1+iswide_unichar, base, 1+iswide_base);
+#endif
+
+			if (base > 0
+#ifdef restrict_subst_decompose
+			    && (! iswide_base && ! iswide_unichar)
+#endif
+			    && (! iswide_base || iswide_unichar)) {
+				/*	wide -> narrow	indicate, append ' ' below
+					wide -> wide	indicate, then return to skip adding ' '
+					narrow -> narrow
+					narrow -> wide	handled in else branch
+				*/
+				indicate_uni (base);
+				if (iswide_base) {
+					/* both (char and subst) are wide */
+					return;
+				}
+			} else if (unichar != (character) UNI_marker) {
+				indicate_narrow ((character) UNI_marker);
+			} else {
+				indicate_narrow (char7bit (UNI_marker));
+			}
 		}
 	}
-	if (iswide (unichar)) {
-		indicate_UTF (' ');
+
+	if (iswide_unichar) {
+		/* fill narrow repl. indication to two columns */
+		indicate_narrow (' ');
 	}
-  }
+  }	/* not utf8_screen */
 }
 
 
@@ -1112,7 +1345,7 @@ put_char (c)
 	} else { /* first UTF-8 byte */
 		if (utfcount > 0) {
 			/* previous character not terminated properly */
-			indicate_illegal_UTF ('');
+			indicate_illegal_UTF ((character) '');
 			utfcount = 0;
 		}
 		if ((c & 0xE0) == 0xC0) {
@@ -1248,10 +1481,12 @@ void
 disp_script (script_colour)
   int script_colour;
 {
-  if (script_colour == default_script) {
-	disp_normal ();
-  } else {
-	disp_colour (script_colour);
+  if (use_script_colour) {
+	if (script_colour == default_script) {
+		disp_normal ();
+	} else {
+		disp_colour (script_colour);
+	}
   }
 }
 
@@ -1262,7 +1497,7 @@ struct colour_mapping {
 	int colour0;
 } colour_table [] =
 {
-#include "colours.h"
+#include "colours.t"
 };
 
 /**
@@ -1323,12 +1558,23 @@ restore_attr ()
 #define dont_debug_width
 
 
+/*
+ * print_line: print line to given screen line
+ */
+void
+print_line (ly, line)
+  int ly;
+  LINE * line;
+{
+  set_cursor (0, ly);
+  put_line (ly, line, 0, True, False);
+}
+
 /*
  * Put_line prints the given line on the standard output.
  * If offset is not zero, printing will start at that x-coordinate.
  * If the FLAG clear_line is True, then the screen line will be cleared
  * when the end of the line has been reached.
-line_print (ly, line) is put_line (ly, line, 0, True, False)
  */
 void
 put_line (ly, line, offset, clear_line, positioning)
@@ -1347,6 +1593,8 @@ put_line (ly, line, offset, clear_line, positioning)
   int offset_stop;
 
   int charwidth;
+  FLAG separate_combining = True;	/* at line beginning, after TAB */
+  FLAG save_combining_mode = combining_mode;
 
   char syntax_marker_was;
   char syntax_marker_new;
@@ -1437,11 +1685,25 @@ put_line (ly, line, offset, clear_line, positioning)
 
   restore_attr ();
 
+  if (offset_start > 0 && * (textp - 1) != '\t') {
+	/* fix display of isolated combinings in case of partial line display */
+	separate_combining = False;
+  }
+
   while (True) {
 	unsigned long unichar;
 	int follow;
 	unsigned long cjkchar;
 
+	if (separate_isolated_combinings) {
+		if (separate_combining) {
+			combining_mode = False;
+			separate_combining = False;
+		} else {
+			combining_mode = save_combining_mode;
+		}
+	}
+
 	if (* textp == '\n') {
 		charwidth = 1;
 		break;
@@ -1462,12 +1724,20 @@ put_line (ly, line, offset, clear_line, positioning)
 		charwidth = cjkscrwidth ((character) * textp, textp, line->text);
 	} else /* Latin-1 */ {
 		unichar = (character) * textp;
-		if (cjk_term || width_data_version & cjk_width_data) {
+		if (cjk_term || cjk_width_data_version) {
 			charwidth = uniscrwidth (unichar, textp, line->text);
 		} else {
 			charwidth = 1;
 		}
 	}
+	if (unassigned_single_width) {
+		if (rxvt_version > 0) {
+			/* handle weird mapping of non-Unicode ranges */
+			if (unichar < 0x80000000) {
+				unichar &= 0x1FFFFF;
+			}
+		}
+	}
 
 	if (count >= offset_stop + 1 - charwidth) {
 		break;
@@ -1488,18 +1758,22 @@ put_line (ly, line, offset, clear_line, positioning)
 		int tab_count = tab (count);
 		int tab_mid = count + (tab_count - count) / 2 + 1;
 
+		if (separate_isolated_combinings) {
+			separate_combining = True;
+		}
+
 		put_uniend ();
 		if (cjk_term) {
 			if ((count & 1) == 1) {
 				count ++;
 				if (cjk_tab_width == 1) {
-					putCJKmark (CJK_TAB_marker);
+					putCJKtab ();
 				} else {
 					__putchar (' ');
 				}
 			}
 			while (count < offset_stop && count < tab_count) {
-				putCJKmark (CJK_TAB_marker);
+				putCJKtab ();
 				count ++;
 				if (cjk_tab_width == 2) {
 					count ++;
@@ -1543,7 +1817,14 @@ put_line (ly, line, offset, clear_line, positioning)
 		ctrldisp_off ();
 		advance_char (& textp);
 		if (white_space (* textp)) {
-			password = NIL_PTR;
+			/*password = NIL_PTR;*/
+			password = strcontains (textp, "assword");
+			if (password) {
+				password += 7;
+				while (* password == ':' || white_space (* password)) {
+					password ++;
+				}
+			}
 		}
 		count += charwidth;
 	} else if (unichar == 0xA0 /* nbsp */) {
@@ -1555,7 +1836,7 @@ put_line (ly, line, offset, clear_line, positioning)
 
 		advance_char (& textp);
 		count += charwidth;
-	} else if (iscontrol (* textp)) {
+	} else if (iscontrol (unichar)) {
 		/* this covers only 1 byte controls; 
 		   GB18030 "C1" controls are handled elsewhere
 		 */
@@ -1563,9 +1844,9 @@ put_line (ly, line, offset, clear_line, positioning)
 
 		if (unichar >= 0x80) {
 			if (unichar == 0x9F) {
-				indicate_UTF ('?');
+				indicate_narrow ('?');
 			} else {
-				indicate_UTF (unichar - 0x20);
+				indicate_narrow (unichar - 0x20);
 			}
 		} else {
 			ctrldisp_on ();
@@ -1575,11 +1856,11 @@ put_line (ly, line, offset, clear_line, positioning)
 
 		restore_attr ();
 
-		textp ++;
+		advance_char (& textp);
 		count ++;
 	} else {
-		FLAG disp_separate_combining;
-		FLAG disp_separate_joining;
+		FLAG disp_separate_combining = False;
+		FLAG disp_separate_joining = False;
 		char syntax_byte_printed;
 
 		syntax_byte_printed = * textp;
@@ -1602,8 +1883,6 @@ put_line (ly, line, offset, clear_line, positioning)
 				}
 			}
 
-			disp_separate_combining = False;
-			disp_separate_joining = False;
 			if (iscombined (unichar, textp, line->text)) {
 			    if (combining_screen && ! combining_mode) {
 				/* enforce separated display */
@@ -1688,6 +1967,8 @@ put_line (ly, line, offset, clear_line, positioning)
 				int charlen;
 				character cjkbytes [5];
 				int cjklen = CJK_len (textp);
+				char * this_textp = textp;
+
 				charlen = cjkencode (cjkchar, cjkbytes);
 				trace_put_line (("%04X %d + %d ", cjkchar, count, charwidth));
 
@@ -1709,15 +1990,15 @@ put_line (ly, line, offset, clear_line, positioning)
 				} else {
 					trace_put_line (("- put %04X ", cjkchar));
 
-					disp_separate_combining = False;
-					if (iscombining (unichar)) {
+					if (iscombined (unichar, this_textp, line->text)) {
 					    if (combining_screen && ! combining_mode) {
 						/* enforce separated display */
 						disp_separate_combining = True;
 						combiningdisp_on ();
-
 						if (charwidth == 2) {
 							put_unichar (0x3000);
+						} else if (isjoined (unichar, this_textp, line->text)) {
+							disp_separate_joining = True;
 						} else {
 							putcharacter (' ');
 						}
@@ -1732,7 +2013,11 @@ put_line (ly, line, offset, clear_line, positioning)
 						combiningdisp_on ();
 					}
 
-					put_cjkcharacter (False, cjkchar, charwidth);
+					if (disp_separate_joining) {
+						put_unjoined (unichar);
+					} else {
+						put_cjkcharacter (False, cjkchar, charwidth);
+					}
 
 					if (disp_separate_combining) {
 						combiningdisp_off ();
@@ -1748,15 +2033,18 @@ put_line (ly, line, offset, clear_line, positioning)
 				count += charwidth;
 			}
 		} else /* ! utf_text && ! cjk_text */ {
-			disp_separate_combining = False;
 			if (mapped_text) {
-				if (iscombining (unichar)) {
+				if (iscombined (unichar, textp, line->text)) {
 				    if (combining_screen && ! combining_mode) {
 					/* enforce separated display */
 					disp_separate_combining = True;
 					combiningdisp_on ();
 
-					putcharacter (' ');
+					if (isjoined (unichar, textp, line->text)) {
+						disp_separate_joining = True;
+					} else {
+						putcharacter (' ');
+					}
 				    } else {
 					if (! combining_screen) {
 						disp_separate_combining = True;
@@ -1769,7 +2057,11 @@ put_line (ly, line, offset, clear_line, positioning)
 			if (mapped_text && unichar == 0xA0 /* nbsp */) {
 				indicate_special ('');
 			} else {
-				putcharacter ((character) * textp);
+				if (disp_separate_joining) {
+					put_unjoined (unichar);
+				} else {
+					putcharacter ((character) * textp);
+				}
 			}
 
 			if (disp_separate_combining) {
@@ -1778,7 +2070,7 @@ put_line (ly, line, offset, clear_line, positioning)
 			}
 
 			textp ++;
-			count ++;
+			count += charwidth;
 		}
 
 		if (dim_HTML && syntax_byte_printed == '>') {
@@ -1786,6 +2078,11 @@ put_line (ly, line, offset, clear_line, positioning)
 		}
 	}
   }
+
+  if (separate_isolated_combinings) {
+	combining_mode = save_combining_mode;
+  }
+
   put_uniend ();
 
   if (script_colour != default_script) {
@@ -1812,7 +2109,7 @@ put_line (ly, line, offset, clear_line, positioning)
 			putcharacter (* textp);
 		}
 	} else if (cjk_term) {
-		putCJKmark (0x300A);	/* 《 */
+		putCJKret (line->return_type);
 	} else if (RET_marker != '\0') {
 		if (PARA_marker != '\0' && paradisp
 		 && line->return_type != lineend_LS
@@ -1862,8 +2159,8 @@ put_line (ly, line, offset, clear_line, positioning)
 		if (RET_marker >= ' ' && (character) RET_marker <= '@') {
 			putmark (RET_marker, NIL_PTR);
 		} else {
-			putCJKmark (0x300A);	/* 《 */
-			/* to assure it fits, scrollbar_width is set to 1 
+			putCJKret (line->return_type);
+			/* to ensure it fits, scrollbar_width is set to 1 
 			   in CJK terminal mode even if scrollbar is switched off */
 			if (cjk_lineend_width == 2) {
 				count ++;
@@ -1926,14 +2223,18 @@ put_line (ly, line, offset, clear_line, positioning)
 
 	/* Clear the rest of the line if clear_line is True */
 	if (clear_line) {
-		set_scrollbar_dirty (ly);
-		if (can_clear_eol) {
+		if (can_erase_chars) {
+			if (XBREAK - count + 1 > 0) {
+				erase_chars (XBREAK - count + 1);
+			}
+		} else if (can_clear_eol) {
 			if (count <= XBREAK) {
 				clear_eol ();
+				set_scrollbar_dirty (ly);
 			}
 		} else {
 			while (count ++ <= XBREAK) {
-				/* clear up to XMAX */
+				/* clear up to scrollbar */
 				putchar (' ');
 			}
 		}
@@ -1975,8 +2276,8 @@ display_line_at (y_pos, line, y_min, y_max, first)
 {
   line = proceed (line, y_pos - y_min);
   if (y_pos >= y_min && y_pos <= y_max) {
-	set_cursor (0, y_pos);
 	if (line == tail) {
+		set_cursor (0, y_pos);
 		clear_wholeline ();
 	} else {
 		if (first == False) {
@@ -1994,12 +2295,12 @@ display_line_at (y_pos, line, y_min, y_max, first)
 			}
 #else
 			if (display_delay > 0) {
-				(void) char_ready_within (display_delay);
+				(void) char_ready_within (display_delay, NIL_PTR);
 			}
 #endif
 #endif
 		}
-		line_print (y_pos, line);
+		print_line (y_pos, line);
 	}
   }
 }
@@ -2037,8 +2338,7 @@ display (y_coord, line, count, new_pos)
 /* with this code, XBREAK must be set to XMAX - 1 */
 /* Print the lines */
   while (line != tail && count -- >= 0) {
-	set_cursor (0, y_coord);
-	line_print (y_coord, line);
+	print_line (y_coord, line);
 	line = line->next;
 	y_coord ++;
   }
