@@ -5,6 +5,7 @@
 
 #include "mined.h"
 #include "io.h"
+#include "termprop.h"
 
 
 FLAG char_on_status_line = False; /* is output active on status line ? */
@@ -15,9 +16,13 @@ FLAG input_active = False;
 |*			Status and prompt buffer			*|
 \*======================================================================*/
 
-static char * lastinbuf;
-static FLAG lastattrib;
-static char status_buf [maxLINE_LEN];
+/* bottom status contents, kept static for redraw */
+static char * bottom_inbuf;
+static FLAG bottom_attrib;
+static char bottom_status_buf [maxLINE_LEN];
+static int bottom_begin_fmt;
+static FLAG bottom_utf8_text;
+
 static char * mnemo = NIL_PTR;
 
 
@@ -40,7 +45,7 @@ void
 print_mark ()
 {
 	reverse_off ();
-	dim_on ();
+	mark_on ();
 	if (cjk_term) {
 		put_cjkchar (encodedchar (0x300A));	/* 《 */
 		if (cjk_lineend_width == 2) {
@@ -50,7 +55,7 @@ print_mark ()
 		put_unichar ((character) RET_marker);
 	}
 	lpos ++;
-	dim_off ();
+	mark_off ();
 	reverse_on ();
 }
 
@@ -260,6 +265,12 @@ print_unichar (unichar, width, char_begin, line_begin)
 
 /**
    Print a string to status line, limited by screen width if limit > 0.
+   int limit: either 0 or XBREAK_bottom
+   FLAG attributes = True enables formatting controls in text string:
+			unicode indicator (cyan background)
+			highlight (red background)
+			reverse
+			plain (normal)
  */
 static
 void
@@ -283,8 +294,8 @@ printlim_string (text, limit, attributes)
 	if (utf8_text) {
 		utf8_info (text, & utfcount, & unichar);
 		charwidth = uniscrwidth (unichar, text, start);
-	} else if (iscontrol (* text)) {
-		if (attributes && * text != '\n') {
+	} else if (iscontrol ((character) * text)) {
+		if (attributes && text >= start + bottom_begin_fmt && * text != '\n') {
 			charwidth = 0;
 		} else {
 			charwidth = 2;
@@ -300,7 +311,7 @@ printlim_string (text, limit, attributes)
 		putmark (SHIFT_marker, UTF_SHIFT_marker);
 		break;
 	}
-	if (lpos > XBREAK_bottom - 1) {
+	if (lpos > XBREAK_bottom - charwidth) {
 		putstring ("\r\n");
 		lpos = 0;
 		pagewrapped = True;
@@ -308,19 +319,20 @@ printlim_string (text, limit, attributes)
 	}
 
 	if (* text == '\n') {
-		print_mark ();
+		print_mark ();	/* also incrementing lpos */
 		text ++;
-	} else if (iscontrol (* text)) {
-		if (attributes) {
+	} else if (iscontrol ((character) * text)) {
+		if (attributes && text >= start + bottom_begin_fmt) {
 			switch (* text) {
-			case '':	unidisp_on ();
+			case '':	unidisp_on ();
 					break;
-			case '':	dim_on ();
+			case '':	emph_on ();
 					menuheader_on ();
 					break;
-			case '':	disp_normal ();
+			case '':	disp_normal ();
 					break;
-			case '':	diagdisp_on ();
+			case '':	disp_normal ();
+					diagdisp_on ();
 					break;
 			}
 			text ++;
@@ -330,7 +342,12 @@ printlim_string (text, limit, attributes)
 			put_unichar (controlchar (* text ++));
 		}
 	} else if (utf8_text) {
-		print_unichar (unichar, charwidth, text, start);
+		if (unichar == 0xA0 && attributes && text >= start + bottom_begin_fmt) {
+			/* skip NO-BREAK SPACE separator */
+		} else {
+			/* print and increment lpos */
+			print_unichar (unichar, charwidth, text, start);
+		}
 		text += utfcount;
 	} else if (cjk_text) {
 		/*int prev_lpos = lpos;*/
@@ -749,7 +766,7 @@ erase1 (inbuf, ptr, all_combined)
 
 	diagdisp_off ();
 	precede_char (& ptr, inbuf);
-	if (iscontrol (* ptr)) {
+	if (iscontrol ((character) * ptr)) {
 	    if (* ptr == '\n' && (cjk_lineend_width == 1 || ! cjk_term)) {
 		/* single-width newline indication */
 		putstring (" \b\b \b");
@@ -769,7 +786,7 @@ erase1 (inbuf, ptr, all_combined)
 				charval = charvalue (ptr);
 			}
 		} else if (iscombined (unicode (charval), ptr, inbuf)) {
-			/* work-around for refresh of base character 
+			/* workaround for refresh of base character 
 			   without or with fewer combining accents */
 			need_refresh = True;
 		}
@@ -915,6 +932,11 @@ input (inbuf, clearfl, term_input)
 	}
      }
 
+     /* special mapping */
+     if (c == ' ' && (keyshift & ctrlshift_mask) == ctrlshift_mask) {
+	c = 0xA0;	/* NO-BREAK SPACE */
+     }
+
      if (command (c) == SNL || command (c) == Popmark) {
 	c = '\r';
      }
@@ -1085,20 +1107,19 @@ rd_bottom_line ()
 {
   set_cursor (0, YMAX);
   diagdisp_on ();
-  if (lastinbuf == NIL_PTR) {
+  if (bottom_inbuf == NIL_PTR) {
 	lpos = 0;
-	if (lastattrib == VALID) {
-		FLAG save_utf8_text = utf8_text;
+	if (bottom_utf8_text && ! utf8_text) {
 		utf8_text = True;
-		printlim_string (status_buf, XBREAK_bottom, True);
-		utf8_text = save_utf8_text;
+		printlim_string (bottom_status_buf, XBREAK_bottom, bottom_attrib == VALID);
+		utf8_text = False;
 	} else {
-		printlim_string (status_buf, XBREAK_bottom, False);
+		printlim_string (bottom_status_buf, XBREAK_bottom, bottom_attrib == VALID);
 	}
   } else {
 	lpos = 0;
-	print_string (status_buf);
-	print_string (lastinbuf);
+	print_string (bottom_status_buf);
+	print_string (bottom_inbuf);
   }
   if (! input_active) {
 	diagdisp_off ();
@@ -1114,6 +1135,12 @@ rd_bottom_line ()
    attrib: True: turn on reverse video on both strings
            False: don't (clear_status)
            VALID: interpret control chars in string for attributes
+			(at beginning) suppress embedding spaces
+		plus printlim_string controls:
+			unicode indicator (cyan background)
+			highlight (red background)
+			reverse
+			plain (normal)
  */
 int
 bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
@@ -1129,19 +1156,34 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   if (inbuf != NIL_PTR) {
 	* inbuf = '\0';
   }
-  lastinbuf = inbuf;
-  lastattrib = attrib;
 
   if (pagewrapped) {
-	status_buf [0] = '\0';
+	bottom_status_buf [0] = '\0';
 	RD ();
 	pagewrapped = False;
   }
 
+  /* set static status buffer (for redrawing):
+     - input buffer
+     - markup attributre
+     - UTF-8 flag
+     - buffer contents
+     - begin of formatting marker
+   */
+  bottom_inbuf = inbuf;
+  bottom_attrib = attrib;
+  bottom_utf8_text = utf8_text;
+  bottom_begin_fmt = 0;
   if (attrib == VALID && s1 && * s1 == '') {
-	build_string (status_buf, "%s%s", unnull (s1), unnull (s2));
+	build_string (bottom_status_buf, "%s%s", unnull (s1), unnull (s2));
+	if (s2 != NIL_PTR) {
+		bottom_begin_fmt = strlen (unnull (s1));
+	}
   } else {
-	build_string (status_buf, " %s%s ", unnull (s1), unnull (s2));
+	build_string (bottom_status_buf, " %s%s ", unnull (s1), unnull (s2));
+	if (s2 != NIL_PTR) {
+		bottom_begin_fmt = strlen (unnull (s1)) + 1;
+	}
   }
 
   if (attrib != False && stat_visible) {
@@ -1159,10 +1201,10 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   }
   if (inbuf == NIL_PTR) {
 	lpos = 0;
-	printlim_string (status_buf, XBREAK_bottom, attrib == VALID);
+	printlim_string (bottom_status_buf, XBREAK_bottom, attrib == VALID);
   } else {
 	lpos = 0;
-	print_string (status_buf);
+	print_string (bottom_status_buf);
 	input_active = True;
 	ret = input (inbuf, statfl, term_input);
 	input_active = False;
@@ -1178,13 +1220,13 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
 	if (proportional) {
 		set_cursor (0, YMAX);
 		diagdisp_on ();
-		if (lastinbuf == NIL_PTR) {
+		if (bottom_inbuf == NIL_PTR) {
 			lpos = 0;
-			printlim_string (status_buf, XBREAK_bottom, attrib == VALID);
+			printlim_string (bottom_status_buf, XBREAK_bottom, attrib == VALID);
 		} else {
 			lpos = 0;
-			print_string (status_buf);
-			print_string (lastinbuf);
+			print_string (bottom_status_buf);
+			print_string (bottom_inbuf);
 		}
 		/* pagewrapped ? diagdisp_off ()? */
 	} else {
@@ -1223,7 +1265,7 @@ get_filename (message, file)
   if (file [0] == '~' && (file [1] == '/' || file [1] == '\0')) {
 	filei = file; filei ++;
 	build_string (file1, "%s%s", envvar ("HOME"), filei);
-	build_string (file, file1);
+	strcpy (file, file1);
   }
 #endif
   return ret;
