@@ -26,8 +26,8 @@
  */
 struct regex {
   union {
-  	char * err_mess;
-  	int * expression;
+	char * err_mess;
+	int * expression;
   } result;
   char status;
   LINE * start_line;
@@ -46,9 +46,11 @@ typedef struct regex REGEX;
 |*		Local function declarations and variables		*|
 \*======================================================================*/
 
+static int compile _((character * pat_start, REGEX * program, FLAG case_ignore));
+static REGEX * make_expression _((char * expr, FLAG case_ignore));
+
 static char * substitute _((LINE * line, LINE * * line_poi, REGEX * program, 
 	int * old_lines_deleted, char * replacement, int * new_lines_inserted));
-static int compile _((character * pat_start, REGEX * program));
 static void search _((char * message, FLAG method));
 static void re_search _((void));
 static void prev_search _((void));
@@ -59,7 +61,6 @@ static int check_string _((REGEX * program, character * str, LINE * from, int *
 static int in_exprlist _((int * exprlist, character * s, int exprlist_length, int opcode));
 static int star _((REGEX * program, char * end_position, char * str, LINE * from, int * expression));
 static LINE * search_it _((REGEX * program, FLAG method));
-static REGEX * make_expression _((char * expr));
 
 
 static FLAG wrapped_search;		/* temporary flag of wrapped search */
@@ -142,7 +143,7 @@ SCURCHAR (method)
   }
   strncat (cbuf, cur_text, cp - cur_text);
 
-  search_for (cbuf, method);
+  search_for (cbuf, method, True);
 }
 
 /*
@@ -211,7 +212,7 @@ RSreverse ()
 {
   if (hop_flag > 0) {
 	hop_flag = 0;
-	search_for (prevexpr, reverse_method (prevmethod));
+	search_for (prevexpr, reverse_method (prevmethod), True);
   } else {
 	do_search (lastprogram, reverse_method (lastmethod));
   }
@@ -244,7 +245,7 @@ get_expression (message)
   if (exp_buf [0] != '\0') {			/* A new expr. was given */
 	copy_string (typed_expression, exp_buf);	/* Save expr. */
 	/* Compile new expression: */
-	if (compile (exp_buf, & program) == ERRORS) {
+	if (compile (exp_buf, & program, True) == ERRORS) {
 		return NIL_REG;
 	}
   }
@@ -263,12 +264,13 @@ get_expression (message)
  */
 static
 REGEX *
-make_expression (expr)
+make_expression (expr, case_ignore)
   char * expr;
+  FLAG case_ignore;
 {
   static REGEX program;		/* Program of expression */
 
-  if (compile (expr, & program) == ERRORS) {
+  if (compile (expr, & program, case_ignore) == ERRORS) {
 	/* Compile new expression */
 	return NIL_REG;
   }
@@ -308,8 +310,6 @@ change (message, global, confirm)
   int previousy = y;
   char c;
   FLAG quit_change;
-  int old_lines_deleted;
-  int new_lines_inserted;
   int lines_changed = 0;
 
   if (viewonly) {
@@ -352,6 +352,8 @@ change (message, global, confirm)
   flush ();
 /* Substitute until end of file */
   do {
+	int old_lines_deleted = 0;
+	int new_lines_inserted;
 	if (line_check (program, searchfrom, line, FORWARD)) {
 	    lines ++;
 	    /* Repeat sub. on this line as long as we find a match */
@@ -374,8 +376,7 @@ change (message, global, confirm)
 				/* update display */
 				display (y, cur_line, last_y - y, y);
 			    } else {
-				set_cursor (0, ly);
-				line_print (ly, line);
+				print_line (ly, line);
 			    }
 			    if (textp == NIL_PTR) {
 				/* substitution failed (too long or no mem) */
@@ -391,6 +392,10 @@ change (message, global, confirm)
 		    textp = substitute (line, & line, 
 					program, & old_lines_deleted, 
 					replacement, & new_lines_inserted);
+#ifdef debug_repl_eol
+			printf ("subst(%d) |%s         ^%s", 
+				old_lines_deleted, line->text, textp);
+#endif
 		    if (textp == NIL_PTR) {
 			/* substitution failed (line too long or no memory) */
 			if (ly > SCREENMAX) {
@@ -403,8 +408,7 @@ change (message, global, confirm)
 				display (y, cur_line, last_y - y, y);
 				reset (top_line, y);	/* reset bot_line */
 			} else {
-				set_cursor (0, ly);
-				line_print (ly, line);
+				print_line (ly, line);
 			}
 			move_to (x, ly);
 			return;
@@ -427,16 +431,19 @@ change (message, global, confirm)
 			reset (top_line, y);	/* reset bot_line */
 #endif
 		} else {
-			set_cursor (0, ly);
-			line_print (ly, line);
+			print_line (ly, line);
 		}
 	    }
 	}
 	if (ly <= SCREENMAX) {
 		ly ++;
 	}
-	line = line->next;
-	searchfrom = line->text;
+	if (old_lines_deleted > 0) {
+		searchfrom = textp;
+	} else {
+		line = line->next;
+		searchfrom = line->text;
+	}
   } while (line != tail && global == VALID && quit == False);
   if (confirm == False && lines_changed) {
 	RDwin ();
@@ -448,7 +455,7 @@ change (message, global, confirm)
 	error ("Pattern not found");
   } else if (lines >= REPORT_CHANGED_LINES || quit) {
 	build_string (mess_buf, "%s %ld substitutions on %ld lines",
-		quit_change ? "(Aborted) " : "", subs, lines);
+		quit_change ? "(Replace aborted) " : "", subs, lines);
 	status_msg (mess_buf);
   } else if (global == NOT_VALID && subs >= REPORT_CHANGED_LINES) {
 	status_line (dec_out (subs), " substitutions");
@@ -510,7 +517,7 @@ substitute (line, line_poi,
 	if (* subp == '&') {		/* Replace the original match */
 		amp = program->start_ptr;
 		ampl = program->start_line;
-/* !!!!!!!!!!!!! Im Abbruchfall hier keine Fehlermeldung !!!!!!!!!!! */
+		/* Do not report an error here if aborting */
 		while (textp < & text_buffer [MAX_CHARS] &&
 			(ampl != program->end_line || amp < program->end_ptr)
 		) {
@@ -562,13 +569,13 @@ substitute (line, line_poi,
   if (program->end_line != program->start_line) {
 	(void) delete_text (line->next, line->next->text, 
 		     program->end_line->next, program->end_line->next->text);
-	* old_lines_deleted = 999;
+	* old_lines_deleted = 9;
   } else {
 	* old_lines_deleted = 0;
   }
 
-/* !!!!!!!!!!! Hier ggf. (falls > 1 newline im Buffer) mehrere Zeilen 
-einfgen; Rckkehrparameter no_of_new_lines einfhren und setzen */
+
+/* Insert replacement string; may be multiple lines */
 
   * new_lines_inserted = 0;
 
@@ -680,7 +687,7 @@ static
 void
 prev_search ()
 {
-  search_for (prevexpr, prevmethod);
+  search_for (prevexpr, prevmethod, True);
 }
 
 static
@@ -717,14 +724,15 @@ search (message, method)
    Search for expression, do not remember it for next search.
  */
 void
-search_for (expr, method)
+search_for (expr, method, case_ignore)
   char * expr;
   FLAG method;
+  FLAG case_ignore;
 {
   register REGEX * program;
 
 /* make the expression */
-  if ((program = make_expression (expr)) == NIL_REG) {
+  if ((program = make_expression (expr, case_ignore)) == NIL_REG) {
 	return;
   }
   do_search (program, method);
@@ -734,14 +742,15 @@ search_for (expr, method)
    Search for expression, remember it for next search.
  */
 void
-search_expr (expr, method)
+search_expr (expr, method, case_ignore)
   char * expr;
   FLAG method;
+  FLAG case_ignore;
 {
   register REGEX * program;
 
 /* make the expression */
-  if ((program = make_expression (expr)) == NIL_REG) {
+  if ((program = make_expression (expr, case_ignore)) == NIL_REG) {
 	return;
   }
 
@@ -758,7 +767,7 @@ search_expr (expr, method)
  * The parameter 'method' must be valid.
  * Special handling is triggered by special values of 'corresponding':
    "/": XML tag matching
-   "#": #if/#else/#endif matching
+   "#": #if/#elif/#else/#endif matching
  */
 void
 search_corresponding (expr, method, corresponding)
@@ -779,7 +788,7 @@ search_corresponding (expr, method, corresponding)
 #endif
 
 /* make the expression */
-  if ((program = make_expression (expr)) == NIL_REG) {
+  if ((program = make_expression (expr, True)) == NIL_REG) {
 	return;
   }
 
@@ -807,7 +816,7 @@ search_corresponding (expr, method, corresponding)
 				|| * program->end_ptr == '>'
 				|| * program->end_ptr == '\0';
 		} else if (* corresponding == '#') {
-			/* #if/#else/#endif matching */
+			/* #if/#elif/#else/#endif matching */
 			char * poi = program->start_ptr;
 			poi ++;
 			while (white_space (* poi) && * poi != '\n' && * poi != '\0') {
@@ -816,7 +825,7 @@ search_corresponding (expr, method, corresponding)
 			if (strisprefix ("if", poi)) {
 				matched = 1;
 				nestingweight = 1;
-			} else if (strisprefix ("els", poi)) {
+			} else if (strisprefix ("el", poi)) {
 				matched = 1;
 				nestingweight = 0;
 			} else if (strisprefix ("end", poi)) {
@@ -843,7 +852,7 @@ search_corresponding (expr, method, corresponding)
 			bracount ++;
 		}
 	} else if (* corresponding == '#') {
-		/* #if/#else/#endif matching */
+		/* #if/#elif/#else/#endif matching */
 		if (bracount == 0 && nestingweight == 0) {
 			bracount = -1;
 		} else if (method == FORWARD) {
@@ -913,7 +922,7 @@ search_it (program, method)
 /* Find the match */
   if ((match_line = match (program, cur_text, method)) == NIL_LINE) {
 	if (quit) {
-		status_msg ("Aborted");
+		status_msg ("Search aborted");
 /*		swallow_dummy_quit_char (); */
 		quit = False;
 	} else {
@@ -1092,6 +1101,11 @@ print_expression (expr)
 #define trace_undent()	
 #endif
 
+
+/*======================================================================*\
+|*			Compile search expression			*|
+\*======================================================================*/
+
 /*
  * Compile compiles the pattern into a more comprehensible form and returns a
  * REGEX structure. If something went wrong, the status field of the structure
@@ -1101,9 +1115,10 @@ print_expression (expr)
  */
 static
 int
-compile (pat_start, program)
+compile (pat_start, program, case_ignore)
   character * pat_start;	/* Pointer to pattern */
   REGEX * program;
+  FLAG case_ignore;
 {
   register int * expression = exp_buffer;
   character * pattern = pat_start;
@@ -1114,8 +1129,7 @@ compile (pat_start, program)
 
   unsigned long low_char;
   unsigned long high_char;
-  int utflen;
-  int utfcount;
+  int charlen;
   character utfbuf [7];
   character * utfpoi;
 
@@ -1165,6 +1179,8 @@ compile (pat_start, program)
 			c = * pattern ++;
 			if (c == 'n') {
 				* expression ++ = NORMAL + '\n';
+			} else if (c == 'r') {
+				* expression ++ = NORMAL + '\r';
 			} else if (c == '0') {
 				* expression ++ = NORMAL + '\0';
 			} else {
@@ -1236,9 +1252,9 @@ compile (pat_start, program)
 			/* Build list */
 			while (low_char <= high_char && low_char != 0) {
 				if (utf8_text) {
-					utflen = utfencode (low_char, utfbuf);
+					charlen = utfencode (low_char, utfbuf);
 				} else if (cjk_text) {
-					utflen = cjkencode (low_char, utfbuf);
+					charlen = cjkencode (low_char, utfbuf);
 				} else {
 					utfbuf [0] = low_char;
 					utfbuf [1] = '\0';
@@ -1283,20 +1299,35 @@ compile (pat_start, program)
 		 */
 		previous (expression) |= BRACKET;
 		break;
-	default :
-		if ((utf8_text || cjk_text) && multichar (c)) {
+	default : {
+		  unsigned long letter;
+		  unsigned long upper;
+		  unsigned long title;
+		  FLAG do_case_ignore;
+
+		  pattern --;	/* reset to begin of character */
+
+		  /* trigger case-insensitive search on small letter */
+		  letter = unicodevalue (pattern);
+		  upper = case_convert (letter, 1);
+		  do_case_ignore = case_ignore && ! no_char (upper) && upper != letter;
+
+		  if (do_case_ignore || ((utf8_text || cjk_text) && multichar (c))) {
+			int charcount;
+
 			/* simulate single-char [u] pattern */
 			acct_field = expression ++;
-			pattern --;	/* reset to begin of character */
 
 			/* determine multi-byte char length */
 			if (utf8_text) {
-				utf8_info (pattern, & utflen, & low_char);
-			} else /* (cjk_text) */ {
-				utflen = CJK_len (pattern);
+				utf8_info (pattern, & charlen, & low_char);
+			} else if (cjk_text) {
+				charlen = CJK_len (pattern);
+			} else {
+				charlen = 1;
 			}
 
-			for (utfcount = 0; utfcount < utflen; utfcount ++) {
+			for (charcount = 0; charcount < charlen; charcount ++) {
 				if (expression >= & exp_buffer [MAX_EXP_SIZE]) {
 					reg_error (too_long);
 					trace_expression ((exp_buffer));
@@ -1305,6 +1336,33 @@ compile (pat_start, program)
 				* expression ++ = * pattern ++;
 			}
 
+			/* make pattern for case-insensitive search */
+			if (do_case_ignore) {
+				char * ec = encode_char (encodedchar (upper));
+				while (* ec) {
+					if (expression >= & exp_buffer [MAX_EXP_SIZE]) {
+						reg_error (too_long);
+						trace_expression ((exp_buffer));
+						return FINE;
+					}
+					* expression ++ = * ec ++;
+				}
+				trace_expression ((exp_buffer));
+				title = case_convert (letter, 2);
+				if (! no_char (title) && title != upper) {
+					char * ec = encode_char (encodedchar (title));
+					while (* ec) {
+						if (expression >= & exp_buffer [MAX_EXP_SIZE]) {
+							reg_error (too_long);
+							trace_expression ((exp_buffer));
+							return FINE;
+						}
+						* expression ++ = * ec ++;
+					}
+					trace_expression ((exp_buffer));
+				}
+			}
+
 			/* Assign length of list in acct field */
 			* acct_field = (int) (expression - acct_field);
 			/* Assign negate and bracket field */
@@ -1314,8 +1372,10 @@ compile (pat_start, program)
 			 * a '*' may be following the list.
 			 */
 			previous (expression) |= BRACKET;
-		} else {
+		  } else {
 			* expression ++ = c + NORMAL;
+			pattern ++;
+		  }
 		}
 	}
 	if (expression == & exp_buffer [MAX_EXP_SIZE]) {
@@ -1327,6 +1387,11 @@ compile (pat_start, program)
   /* NOTREACHED */
 }
 
+
+/*======================================================================*\
+|*			Invoke search					*|
+\*======================================================================*/
+
 /*
  * Match gets as argument the program, pointer to place in current line to
  * start from and the method to search for (either FORWARD or REVERSE).
@@ -1377,7 +1442,11 @@ match (program, string, method)
   do {
 	line = (method == FORWARD) ? line->next : line->prev;
 	if (line->text == NIL_PTR) {	/* header/tail */
-		status_msg ("Search wrapped around end of file");
+		if (method == FORWARD) {
+			status_uni (">>>>>>>> Search wrapped around end of file >>>>>>>>");
+		} else {
+			status_uni ("<<<<<<<< Search wrapped around end of file <<<<<<<<");
+		}
 		wrapped_search = True;
 		continue;
 	}
@@ -1498,11 +1567,7 @@ check_string (program, string, from_line, expression)
   trace_indent ();
 
 /* Loop until end of string or end of expression */
-  while (quit == False && ! (* expression & DONE)
-/*		&& * string != '\n'	*/
-		&& * string != '\0'
-	)
-  {
+  while (quit == False && ! (* expression & DONE) && * string != '\0') {
 	trace_search (("check [%d] %s", (char *) string - from_line->text, string));
 	trace_expression ((expression));
 
@@ -1539,12 +1604,14 @@ check_string (program, string, from_line, expression)
 		/*	advance_char (& string);	*/
 		/*	string ++;	???	*/
 		} else {
-			if (* string != patternchar) {
+			if (! (* string == patternchar
+			       || (* string == '\n' && patternchar == '\r' && from_line->return_type == lineend_CRLF)
+			      )) {
 				trace_undent ();
 				return NO_MATCH;
 			}
-			if (patternchar == '\n') {
-			/* handle multi-line matching */
+			if (patternchar == '\n' || patternchar == '\r') {
+			    /* handle multi-line matching */
 			    if (from_line->return_type == lineend_NONE
 				|| from_line->return_type == lineend_NUL
 				)
